from typing import Optional, Union, Dict, Any
import os
import asyncio
import random
import urllib.parse
from email.message import Message
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from io import BytesIO
import ssl
from pathlib import Path
from urllib.parse import quote, urlencode, urlparse
import urllib3
import aiofiles
import requests
import backoff
from requests.auth import HTTPBasicAuth
from requests.exceptions import HTTPError
from requests.exceptions import Timeout as RequestTimeoutException
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
from duckduckgo_search import DDGS
from duckduckgo_search.exceptions import (
    DuckDuckGoSearchException,
    RatelimitException,
    TimeoutException,
)
import primp
import aiohttp
from aiohttp import BasicAuth
import httpx
from bs4 import BeautifulSoup as bs
from lxml import html, etree
from navconfig.logging import logging
from proxylists.proxies import (
    FreeProxy,
    Oxylabs,
    Decodo,
    Geonode
)
from ..utils import cPrint, SafeDict

from ..conf import (
    HTTPCLIENT_MAX_SEMAPHORE,
    HTTPCLIENT_MAX_WORKERS,
    GOOGLE_SEARCH_API_KEY,
    GOOGLE_SEARCH_ENGINE_ID
)
from .dataframes import PandasDataframe
class ComponentError(Exception):
    """Base class for component errors."""
    pass
from .credentials import CredentialsInterface


logging.getLogger("urllib3").setLevel(logging.WARNING)
urllib3.disable_warnings()
logging.getLogger("httpx").setLevel(logging.WARNING)
logging.getLogger("httpcore").setLevel(logging.WARNING)
logging.getLogger("aiohttp").setLevel(logging.WARNING)
logging.getLogger("rquest").setLevel(logging.INFO)


ua = [
    # Chrome - Desktop (Windows)
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
    # Chrome - Desktop (Mac)
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",  # noqa
    # Safari - Desktop (Mac)
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.0 Safari/605.1.15",  # noqa
    # Firefox - Desktop (Windows)
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:109.0) Gecko/20100101 Firefox/118.0",
    # Edge - Desktop (Windows)
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36 Edg/118.0.2088.46",  # noqa
    # Chrome - Mobile (Android)
    "Mozilla/5.0 (Linux; Android 13; SM-G991B) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Mobile Safari/537.36",  # noqa
    # Safari - Mobile (iOS)
    "Mozilla/5.0 (iPhone; CPU iPhone OS 16_6 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.6 Mobile/15E148 Safari/604.1",  # noqa
    # Samsung Internet - Mobile (Android)
    "Mozilla/5.0 (Linux; Android 13; SAMSUNG SM-G991B) AppleWebKit/537.36 (KHTML, like Gecko) SamsungBrowser/21.0 Chrome/118.0.0.0 Mobile Safari/537.36",  # noqa
    # Firefox - Mobile (Android)
    "Mozilla/5.0 (Android 13; Mobile; rv:118.0) Gecko/118.0 Firefox/118.0",
    # Opera - Desktop (Windows)
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36 OPR/104.0.0.0"  # noqa
    # Firefox - Desktop (Linux)
    "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:133.0) Gecko/20100101 Firefox/133.0",
    # Chrome - Desktop (Linux)
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.0.0 Safari/537.36",
    # Other:
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",  # noqa
]

UA_LIST = ua

mobile_ua = [
    "Mozilla/5.0 (Linux; Android 4.2.1; en-us; Nexus 5 Build/JOP40D) AppleWebKit/535.19 (KHTML, like Gecko) Chrome/18.0.1025.166 Mobile Safari/535.19",  # noqa
    'Mozilla/5.0 (iPhone; CPU iPhone OS 12_0 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.0 Mobile/15E148 Safari/604.1',  # noqa
    'Mozilla/5.0 (Linux; Android 9; SM-G973F) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/80.0.3987.119 Mobile Safari/537.36',  # noqa
    'Mozilla/5.0 (Linux; Android 8.0.0; Pixel 2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/79.0.3945.93 Mobile Safari/537.36',  # noqa
    'Mozilla/5.0 (Linux; Android 10; HUAWEI VOG-L29) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/81.0.4044.138 Mobile Safari/537.36',  # noqa
    'Mozilla/5.0 (iPad; CPU OS 13_2_3 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.0 Mobile/15E148 Safari/604.1',  # noqa
]

impersonates = (
    "chrome_100", "chrome_101", "chrome_104", "chrome_105", "chrome_106", "chrome_107",
    "chrome_108", "chrome_109", "chrome_114", "chrome_116", "chrome_117", "chrome_118",
    "chrome_119", "chrome_120", "chrome_123", "chrome_124", "chrome_126", "chrome_127",
    "chrome_128", "chrome_129", "chrome_130", "chrome_131",
    "safari_ios_16.5", "safari_ios_17.2", "safari_ios_17.4.1", "safari_ios_18.1.1",
    "safari_15.3", "safari_15.5", "safari_15.6.1", "safari_16", "safari_16.5",
    "safari_17.0", "safari_17.2.1", "safari_17.4.1", "safari_17.5",
    "safari_18", "safari_18.2",
    "safari_ipad_18",
    "edge_101", "edge_122", "edge_127", "edge_131",
    "firefox_109", "firefox_117", "firefox_128", "firefox_133",
)  # fmt: skip

impersonates_os = ("android", "ios", "linux", "macos", "windows")

valid_methods = ['GET', 'POST', 'PUT', 'DELETE', 'PATCH', 'HEAD', 'OPTIONS']

def bad_gateway_exception(exc):
    """Check if the exception is a 502 Bad Gateway error."""
    return isinstance(exc, httpx.HTTPStatusError) and exc.response.status_code == 502

class HTTPService(CredentialsInterface, PandasDataframe):
    """
    HTTPService.

    Overview

            Interface for making connections to HTTP services.
    """
    accept: str = "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9"  # noqa

    def __init__(self, *args, **kwargs):
        self.url: str = kwargs.get("url", None)
        self.accept: str = kwargs.get(
            "accept",
            self.accept
        )
        self.use_proxy: bool = kwargs.pop("use_proxy", False)
        self.proxy_type: str = kwargs.pop('proxy_type', 'oxylabs')
        self._free_proxy: bool = kwargs.pop('use_free_proxy', True)
        self._proxies: list = []
        self.rotate_ua: bool = kwargs.pop("rotate_ua", False)
        self.use_async: bool = bool(kwargs.pop("use_async", True))
        self.google_api_key: str = kwargs.pop('google_api_key', GOOGLE_SEARCH_API_KEY)
        self.google_cse: str = kwargs.pop('google_cse', GOOGLE_SEARCH_ENGINE_ID)
        self.as_binary: bool = kwargs.pop('as_binary', False)
        self.download: bool = kwargs.pop('download', False)
        self.timeout: int = 30
        self.headers: dict = kwargs.get('headers', {})
        self.auth: dict = {}
        self.auth_type: str = kwargs.get('auth_type', None)
        self.token_type: str = "Bearer"
        self._debug: bool = kwargs.get('debug', False)
        self._user, self._pwd = None, None
        self.method: str = kwargs.get("method", "get")
        self._default_parser: str = kwargs.pop('bs4_parser', 'html.parser')
        self.parameters = {}
        self._logger = logging.getLogger(__name__)
        if self.rotate_ua is True:
            self._ua = random.choice(ua)
        else:
            self._ua: str = ua[0]
        self.headers = {
            "Accept": self.accept,
            "Accept-Encoding": "gzip, deflate",
            "DNT": "1",
            "Connection": "keep-alive",
            "Upgrade-Insecure-Requests": "1",
            "User-Agent": self._ua,
            **self.headers,
        }
        # potentially cookies to be used by request.
        self.cookies = kwargs.get('cookies', {})

        # other arguments:
        self.arguments = kwargs
        # Executor:
        self._executor = ThreadPoolExecutor(
            max_workers=int(HTTPCLIENT_MAX_WORKERS)
        )
        self._semaphore = asyncio.Semaphore(
            int(HTTPCLIENT_MAX_SEMAPHORE)
        )
        super().__init__(*args, **kwargs)

    def add_metric(self, key: str, value: Any) -> None:
        """
        Stub method for adding metrics.
        Override in subclasses if needed.
        """
        pass

    async def get_proxies(
        self,
        session_time: float = 0.40,
        free_proxy: bool = False
    ):
        """
        Asynchronously retrieves a list of free proxies.
        TODO: SELECT or rotate the free/paid proxies.
        """
        if self._free_proxy is True:
            return await FreeProxy().get_list()
        else:
            if self.proxy_type == 'decodo':
                return await Decodo().get_list()
            elif self.proxy_type == 'oxylabs':
                return await Oxylabs(
                    session_time=session_time,
                    timeout=10
                ).get_list()
            elif self.proxy_type == 'geonode':
                return await Geonode().get_list()
            else:
                return []

    async def refresh_proxies(self):
        """
        Asynchronously refreshes the list of proxies if proxy usage is enabled.
        """
        if self.use_proxy is True:
            self._proxies = await self.get_proxies()

    def build_url(self, url, queryparams: str = "", args=None):
        """
        Constructs a full URL with optional query parameters and arguments.

        :param url: The base URL to be formatted.
        :param queryparams: Additional query parameters to be appended to the URL.
        :param args: Arguments to format within the URL.
        :return: The fully constructed URL.
        """
        url = str(url).format_map(SafeDict(**self._variables))
        if args:
            u = url.format(**args)
        else:
            u = url
        if queryparams:
            if "?" in u:
                full_url = u + "&" + queryparams
            else:
                full_url = u + "?" + queryparams
        else:
            full_url = u
        logging.debug(
            f"Resource URL: {full_url!s}"
        )
        return full_url

    def extract_host(self, url):
        parsed_url = urlparse(url)
        return parsed_url.netloc

    async def session(
        self,
        url: str,
        method: str = "get",
        data: dict = None,
        cookies: dict = None,
        headers: dict = None,
        use_json: bool = False,
        follow_redirects: bool = False,
        use_proxy: bool = False,
        accept: str = None,
        return_response: bool = False
    ):
        """
        Asynchronously sends an HTTP request using HTTPx.

        :param url: The URL to send the request to.
        :param method: The HTTP method to use (e.g., 'GET', 'POST').
        :param data: The data to send in the request body.
        :param use_json: Whether to send the data as JSON.
        :param cookies: A dictionary of cookies to send with the request.
        :param headers: A dictionary of headers to send with the request.
        :return: A tuple containing the result and any error information.
        """
        result = []
        error = {}
        auth = None
        proxies = None
        if accept is not None:
            self.headers["Accept"] = accept
        else:
            self.headers["Accept"] = self.accept
        if use_proxy is True:
            self._proxies = await self.get_proxies()
            if len(self._proxies) == 1:
                proxies = self._proxies[0]
                if not proxies.startswith('http'):
                    proxies = f"http://{proxies}"
            elif len(self._proxies) > 1:
                proxy = random.choice(self._proxies)
                if not proxy.startswith('http'):
                    proxies = f"http://{proxy}"
                else:
                    proxies = proxy
            else:
                self._proxies = None
        if self.credentials:
            if "apikey" in self.auth:
                self.headers[
                    "Authorization"
                ] = f"{self.token_type} {self.auth['apikey']}"
            elif self.auth_type == "api_key":
                self.headers = {**self.headers, **self.credentials}
            elif self.auth_type == "key":
                url = self.build_url(
                    url, args=self.arguments, queryparams=urlencode(self.credentials)
                )
            elif self.auth_type in ["basic", "auth", "user"]:
                auth = (self.credentials["username"], self.credentials["password"])
        elif self._user and self.auth_type == "basic":
            auth = (self._user, self._pwd)
        cPrint(
            f"HTTP: Connecting to {url} using {method}",
            level="DEBUG"
        )
        if self.download is True:
            self.headers["Accept"] = "application/octet-stream"
            self.headers["Content-Type"] = "application/octet-stream"
            if self.use_streams is True:
                self.headers["Transfer-Encoding"] = "chunked"
        headers = self.headers
        if headers is not None and isinstance(headers, dict):
            headers = {**self.headers, **headers}
        timeout = httpx.Timeout(self.timeout)
        args = {"timeout": timeout, "headers": headers, "cookies": cookies}
        if auth is not None:
            args["auth"] = auth
        if proxies:
            if isinstance(proxies, dict):
                args['mounts'] = proxies
            else:
                args["proxies"] = proxies
        # if self._debug is True:
        #    self.add_metric("HEADERS", headers)
        if proxies is not None:
            self.add_metric('Proxies', proxies)
        self.add_metric('URL', url)
        self.add_metric('METHOD', method)
        req_args = {
            "method": method.upper(),
            "url": url,
            "follow_redirects": follow_redirects,
            "json" if use_json else "data": data
        }
        # Process the response
        try:
            if hasattr(self, "_client"):
                # Use a client without context manager to keep the session alive
                # Remember to call `await self._client.aclose()` manually
                response = await self._client.request(**req_args)
            else:
                async with httpx.AsyncClient(**args) as client:
                    response = await client.request(**req_args)

            result, error = await self.process_response(response, url)

            if return_response:
                return response, result, error

        except httpx.HTTPError as e:
            error = str(e)

        return (result, error)

    async def async_request(
        self,
        url,
        method: str = 'GET',
        data: dict = None,
        use_json: bool = False,
        use_proxy: bool = False,
        accept: Optional[str] = None
    ):
        """
        Asynchronously sends an HTTP request using aiohttp.

        :param url: The URL to send the request to.
        :param method: The HTTP method to use (e.g., 'GET', 'POST').
        :param data: The data to send in the request body.
        :param use_json: Whether to send the data as JSON.
        :param use_proxy: force proxy usage.
        :return: A tuple containing the result and any error information.
        """
        result = []
        error = {}
        auth = None
        proxy = None

        if use_proxy is True:
            self._proxies = await self.get_proxies()
        if self._proxies:
            proxy = random.choice(self._proxies)
            self.add_metric("Proxies", proxy)
        if self.credentials:
            if "apikey" in self.auth:
                self.headers[
                    "Authorization"
                ] = f"{self.token_type} {self.auth['apikey']}"
            elif self.auth_type == "api_key":
                self.headers = {**self.headers, **self.credentials}
            elif self.auth_type == "key":
                url = self.build_url(
                    url,
                    args=self.arguments,
                    queryparams=urlencode(self.credentials)
                )
            elif self.auth_type in ["basic", "auth", "user"]:
                auth = BasicAuth(
                    self.credentials["username"],
                    self.credentials["password"]
                )
        elif "apikey" in self.auth:
            self.headers["Authorization"] = f"{self.token_type} {self.auth['apikey']}"
        elif self.auth:
            token_type, token = list(self.auth.items())[0]
            self.headers["Authorization"] = f"{token_type} {token}"
        elif self._user and self.auth_type == "basic":
            auth = BasicAuth(self._user, self._pwd)
        cPrint(
            f"HTTP: Connecting to {url} using {method}",
            level="DEBUG"
        )
        if self._debug is True:
            self.add_metric("HEADERS", self.headers)
        self.add_metric("URL", url)
        self.add_metric("METHOD", method)
        if auth is not None:
            args = {"auth": auth}
        else:
            args = {}
        if accept is not None:
            self.headers["Accept"] = accept
        else:
            self.headers["Accept"] = self.accept
        if self.download is True:
            self.headers["Accept"] = "application/octet-stream"
            self.headers["Content-Type"] = "application/octet-stream"
            if hasattr(self, "use_streams"):
                self.headers["Transfer-Encoding"] = "chunked"
                args["stream"] = True
        timeout = aiohttp.ClientTimeout(total=self.timeout)
        async with aiohttp.ClientSession(
            headers=self.headers, timeout=timeout, auth=auth
        ) as session:
            try:
                if use_json is True:
                    async with session.request(
                        method.upper(), url, json=data, proxy=proxy, **args
                    ) as response:
                        result, error = await self.process_response(response, url)
                else:
                    async with session.request(
                        method.upper(), url, data=data, proxy=proxy, **args
                    ) as response:
                        # Process the response
                        result, error = await self.process_response(response, url)
            except aiohttp.ClientError as e:
                error = str(e)
        return (result, error)
    async def evaluate_error(
        self, response: Union[str, list], message: Union[str, list, dict]
    ) -> tuple:
        """evaluate_response.

        Check Response status and available payloads.
        Args:
            response (_type_): _description_
            url (str): _description_

        Returns:
            tuple: _description_
        """
        if isinstance(response, list):
            # a list of potential errors:
            for msg in response:
                if message in msg:
                    return True
        if isinstance(response, dict) and "errors" in response:
            errors = response["errors"]
            if isinstance(errors, list):
                for error in errors:
                    try:
                        if message in error:
                            return True
                    except TypeError:
                        if message == error:
                            return True
            else:
                if message == errors:
                    return True
        else:
            if message in response:
                return True
        return False

    async def process_response(self, response, url: str) -> tuple:
        """
        Processes the response from an HTTP request.

        :param response: The response object from aiohttp.
        :param url: The URL that was requested.
        :return: A tuple containing the processed result and any error information.
        """
        error = None
        result = None
        # Process the response
        status = self.response_status(response)

        if status >= 400:
            # Evaluate response body and headers.
            print(" == ERROR Headers == ")
            print(f"{response.headers}")
            content_type = response.headers.get("Content-Type", "").lower()
            if "application/json" in content_type:
                message = await self.response_json(response)
            elif "text/" in content_type:
                message = await self.response_text(response)
            elif "X-Error" in response.headers:
                message = response.headers["X-Error"]
            else:
                # Fallback to a unified read method for the raw body content
                message = await self.response_read(response)

            # Combine response headers and body for enriched logging
            error_context = {
                "status": status,
                "reason": await self.response_reason(response),
                "headers": response.headers,
                "body": message
            }

            # Log the detailed error context
            self._logger.error(f"Error: {error_context}")

            # Additional error handling or custom evaluation based on status
            if hasattr(self, 'no_errors'):
                for key, msg in self.no_errors.items():
                    if int(key) == status and await self.evaluate_error(message, msg):
                        return response, status

            # Raise an exception if error handling does not continue
            raise ConnectionError(f"HTTP Error {status}: {message!s}")
        else:
            if self.download is True:
                filename = os.path.basename(url)
                # Get the filename from the response headers, if available
                content_disposition = response.headers.get("content-disposition")
                if content_disposition:
                    msg = Message()
                    msg["Content-Disposition"] = response.headers.get("content-disposition")
                    filename = msg.get_param("filename", header="Content-Disposition")
                    utf8_filename = msg.get_param("filename*", header="Content-Disposition")
                    if utf8_filename:
                        _, utf8_filename = utf8_filename.split("''", 1)
                        filename = urllib.parse.unquote(utf8_filename)
                if "{filename}" in str(self.filename):
                    self.filename = str(self.filename).format_map(
                        SafeDict(filename=filename)
                    )
                if "{" in str(self.filename):
                    self.filename = str(self.filename).format_map(
                        SafeDict(**self.arguments)
                    )
                if isinstance(self.filename, str):
                    self.filename = Path(self.filename)
                # Saving File in Directory:
                total_length = response.headers.get("Content-Length")
                self._logger.info(
                    f"HTTPClient: Saving File {self.filename}, size: {total_length}"
                )
                pathname = self.filename.parent.absolute()
                if not pathname.exists():
                    # Create a new directory
                    pathname.mkdir(parents=True, exist_ok=True)
                transfer = response.headers.get("transfer-encoding", None)
                if transfer is None:
                    chunk_size = int(total_length)
                else:
                    chunk_size = 8192
                # Asynchronous file writing
                if self.filename.exists() and self.filename.is_file():
                    overwrite = self.destination.get("overwrite", True)
                    if overwrite is False:
                        self._logger.warning(
                            f"HTTPClient: File Already exists: {self.filename}"
                        )
                        # Filename already exists
                        result = self.filename
                        return result, error
                    else:
                        self._logger.warning(
                            f"HTTPClient: Overwriting File: {self.filename}"
                        )
                        # Delete the file before downloading again.
                        try:
                            self.filename.unlink()
                        except Exception as e:
                            self._logger.warning(
                                f"HTTPClient: Error Deleting File: {self.filename}, {e}"
                            )
                if hasattr(self, "use_streams") and self.use_streams is True:
                    async with aiofiles.open(self.filename, "wb") as file:
                        async for chunk in response.content.iter_chunked(chunk_size):
                            await file.write(chunk)
                else:
                    with open(self.filename, "wb") as fp:
                        try:
                            fp.write(await self.response_read(response))
                        except Exception:
                            pass
                self._logger.debug(
                    f"Filename Saved Successfully: {self.filename}"
                )
                result = self.filename
            else:
                if self.accept == 'application/octet-stream':
                    data = await self.response_read(response)
                    buffer = BytesIO(data)
                    buffer.seek(0)
                    result = buffer
                elif self.accept in ('text/html'):
                    result = await self.response_read(response)
                    try:
                        # html parser for lxml
                        self._parser = html.fromstring(result)
                        # BeautifulSoup parser
                        self._bs = bs(response.text, self._default_parser)
                        result = self._bs
                    except Exception as e:
                        error = e
                elif self.accept in ('application/xhtml+xml', 'application/xml'):
                    result = await self.response_read(response)
                    try:
                        self._parser = etree.fromstring(result)
                    except etree.XMLSyntaxError:
                        self._parser = html.fromstring(result)
                    except Exception as e:
                        error = e
                elif self.accept == "application/json":
                    try:
                        result = await self.response_json(response)
                    except Exception as e:
                        logging.warning(e)
                        # is not an json, try first with beautiful soup:
                        try:
                            self._bs = bs(
                                await self.response_text(response),
                                self._default_parser
                            )
                            result = self._bs
                        except Exception:
                            error = e
                elif self.as_binary is True:
                    result = await self.response_read(
                        response
                    )
                else:
                    result = await self.response_text(
                        response
                    )
        return result, error

    async def request(
        self,
        url: str,
        method: str = "GET",
        data: dict = None,
        use_proxy: bool = False,
        accept: Optional[str] = None
    ) -> tuple:
        """
        Sends an HTTP request using the requests library.

        :param url: The URL to send the request to.
        :param method: The HTTP method to use (e.g., 'GET', 'POST').
        :param data: The data to send in the request body.
        :return: A tuple containing the result and any error information.
        """
        result = []
        error = {}
        auth = None
        proxies = None
        if use_proxy is True:
            self._proxies = await self.get_proxies()
        if self._proxies:
            proxy = random.choice(self._proxies)
            proxies = {"http": proxy, "https": proxy, "ftp": proxy}
            self.add_metric("Proxies", proxies)
        if self.credentials:
            if "apikey" in self.auth:
                self.headers[
                    "Authorization"
                ] = f"{self.token_type} {self.auth['apikey']}"
            elif self.auth_type == "api_key":
                self.headers = {**self.headers, **self.credentials}
            elif self.auth_type == "key":
                url = self.build_url(
                    url, args=self.arguments, queryparams=urlencode(self.credentials)
                )
            elif self.auth_type == "basic":
                auth = HTTPBasicAuth(
                    self.credentials["username"], self.credentials["password"]
                )
            else:
                auth = HTTPBasicAuth(
                    self.credentials["username"], self.credentials["password"]
                )
        elif self._user and self.auth_type == "basic":
            auth = HTTPBasicAuth(self._user, self._pwd)
        cPrint(f"HTTP: Connecting to {url} using {method}", level="DEBUG")
        self.add_metric("URL", url)
        self.add_metric("METHOD", method)
        if auth is not None:
            args = {"auth": auth, "verify": False}
        else:
            args = {}
        if accept is not None:
            self.headers["Accept"] = accept
        else:
            self.headers["Accept"] = self.accept
        if self.download is True:
            self.headers["Accept"] = "application/octet-stream"
            self.headers["Content-Type"] = "application/octet-stream"
            if hasattr(self, "use_streams"):
                self.headers["Transfer-Encoding"] = "chunked"
                args["stream"] = True
        if self._debug is True:
            self.add_metric("HEADERS", self.headers)
        args["headers"] = self.headers
        args["timeout"] = self.timeout
        args["proxies"] = proxies
        method = method.lower()
        if method == "get":
            my_request = partial(requests.get, **args)
        elif method == "post":
            my_request = partial(requests.post, data=data, **args)
        elif method == "put":
            my_request = partial(requests.put, data=data, **args)
        elif method == "delete":
            my_request = partial(requests.delete, data=data, **args)
        elif method == "patch":
            my_request = partial(requests.patch, data=data, *args)
        else:
            my_request = partial(requests.post, data=data, **args)
        try:
            # making request
            async with self._semaphore:
                loop = asyncio.get_running_loop()
                future = loop.run_in_executor(self._executor, my_request, url)
                result, error = await self.process_request(future, url)
            if error:
                if isinstance(error, BaseException):
                    raise error
                else:
                    raise ComponentError(f"{error!s}")
            return (result, error)
        except requests.exceptions.ReadTimeout as err:
            self._logger.warning(f"Timeout Error: {err!r}")
            # TODO: retrying
            raise ComponentError(f"Timeout: {err}") from err
        except Exception as err:
            self._logger.exception(str(err), stack_info=True)
            raise ComponentError(f"Error: {err}") from err

    async def process_request(self, future, url: str):
        """
        Processes the result of an asynchronous HTTP request.

        :param future: The future object representing the asynchronous operation.
        :param url: The URL that was requested.
        :return: A tuple containing the result and any error information.
        """
        # getting the result, based on the Accept logic
        error = None
        result = None
        loop = asyncio.get_running_loop()
        asyncio.set_event_loop(loop)
        done, _ = await asyncio.wait([future], return_when=asyncio.FIRST_COMPLETED)
        for f in done:
            response = f.result()
            # for response in await asyncio.gather(*future):
            # Check for HTTP errors
            try:
                response.raise_for_status()
            except HTTPError as http_err:
                # Handle HTTP errors here
                error = http_err
                # Log the error or perform other error handling
                self._logger.error(f"HTTP error occurred: {http_err}")
                # You can choose to continue, break, or return based on your logic
                continue
            try:
                if self.download is True:
                    # Filename:
                    filename = os.path.basename(url)
                    # Get the filename from the response headers, if available
                    content_disposition = response.headers.get("content-disposition")
                    if content_disposition:
                        _, params = content_disposition.split(";")
                        try:
                            key, value = params.strip().split("=")
                            if key == "filename":
                                filename = value.strip("'\"")
                        except ValueError:
                            pass
                    if "{filename}" in str(self.filename):
                        self.filename = str(self.filename).format_map(
                            SafeDict(filename=filename)
                        )
                    if "{" in str(self.filename):
                        self.filename = str(self.filename).format_map(
                            SafeDict(**self.arguments)
                        )
                    if isinstance(self.filename, str):
                        self.filename = Path(self.filename)
                    # Saving File in Directory:
                    total_length = response.headers.get("Content-Length")
                    self._logger.info(
                        f"HTTPClient: Saving File {self.filename}, size: {total_length}"
                    )
                    pathname = self.filename.parent.absolute()
                    if not pathname.exists():
                        # Create a new directory
                        pathname.mkdir(parents=True, exist_ok=True)
                    response.raise_for_status()
                    transfer = response.headers.get("transfer-encoding", None)
                    if transfer is None:
                        chunk_size = int(total_length)
                    else:
                        chunk_size = 8192
                    if self.filename.exists() and self.filename.is_file():
                        overwrite = self.destination.get("overwrite", True)
                        if overwrite is False:
                            self._logger.warning(
                                f"HTTPClient: File Already exists: {self.filename}"
                            )
                            # Filename already exists
                            result = self.filename
                            continue
                        else:
                            self._logger.warning(
                                f"HTTPClient: Overwriting File: {self.filename}"
                            )
                            # Delete the file before downloading again.
                            try:
                                self.filename.unlink()
                            except Exception as e:
                                self._logger.warning(
                                    f"HTTPClient: Error Deleting File: {self.filename}, {e}"
                                )
                    with open(self.filename, "wb") as fp:
                        try:
                            for chunk in response.iter_content(chunk_size=chunk_size):
                                fp.write(chunk)
                            fp.flush()
                        except Exception:
                            pass
                    self._logger.debug(f"Filename Saved Successfully: {self.filename}")
                    result = self.filename
                elif self.accept in ("text/html"):
                    result = response.content  # Get content of the response as bytes
                    try:
                        # html parser for lxml
                        self._parser = html.fromstring(result)
                        # BeautifulSoup parser

                        self._bs = bs(response.text, self._default_parser)
                        result = self._bs
                    except Exception as e:
                        error = e
                elif self.accept in ("application/xhtml+xml", "application/xml"):
                    result = response.content  # Get content of the response as bytes
                    try:
                        self._parser = etree.fromstring(result)
                    except Exception as e:
                        error = e
                elif self.accept == "application/json":
                    try:
                        result = response.json()
                    except Exception as e:
                        logging.error(e)
                        # is not an json, try first with beautiful soup:
                        try:
                            self._bs = bs(response.text, self._default_parser)
                            result = self._bs
                        except Exception:
                            error = e
                else:
                    result = response.text
            except requests.exceptions.ProxyError as err:
                raise ComponentError(f"Proxy Connection Error: {err!r}") from err
            except requests.ReadTimeout as err:
                return (result, err)
            except requests.exceptions.HTTPError as e:
                # Log the error or perform other error handling
                self._logger.error(f"HTTP error occurred: {error}")
                raise ComponentError(f"HTTP Error: {error!r}, ex: {e!s}") from e
            except Exception as e:
                logging.exception(e)
                return (result, e)
        # returning results
        return (result, error)

    @staticmethod
    async def response_read(response):
        if hasattr(response, 'aread'):
            return await response.aread()

        return await response.read()

    @staticmethod
    async def response_json(response):
        if asyncio.iscoroutinefunction(response.json):
            return await response.json()

        return response.json()

    @staticmethod
    def response_status(response):
        if hasattr(response, 'status_code'):
            return response.status_code

        return response.status

    @staticmethod
    async def response_text(response):
        if asyncio.iscoroutinefunction(response.text):
            return await response.text()

        return response.text

    @staticmethod
    async def response_reason(response):
        # Attempt to retrieve `reason`, `reason_phrase`, or fallback to an empty string
        reason = getattr(response, 'reason', getattr(response, 'reason_phrase', b''))

        return f"{reason!s}"

    @backoff.on_exception(
        backoff.expo,
        (httpx.HTTPStatusError, httpx.TimeoutException),  # Catch HTTP errors and timeouts
        max_tries=3,
        max_time=120,
        jitter=backoff.full_jitter,
        on_backoff=lambda details: print(
            f"Retrying HTTP Get: attempt {details['tries']} after {details['wait']:0.2f}s"
        ),
    )
    async def _get(
        self,
        url: str,
        cookies: httpx.Cookies = None,
        params: Dict[str, Any] = None,
        headers: Dict[str, str] = None,
        timeout: Union[int, float] = 30.0,
        use_proxy: bool = True,
        free_proxy: bool = False,
        connect_timeout: Union[int, float] = 5.0,
        read_timeout: Union[int, float] = 20.0,
        write_timeout: Union[int, float] = 5.0,
        pool_timeout: Union[int, float] = 20.0,
        num_retries: int = 2,
        **kwargs
    ) -> Dict[str, Any]:
        """
        Make an asynchronous HTTP GET request, returning the response object.

        Args:
            url (str): The URL to send the GET request to.
            cookies (httpx.Cookies): Cookies to include in the request.
            params (dict): Dictionary of query parameters to include in the URL.

        Returns:
            Response: The response object from the httpx.
        """
        proxies = None
        if use_proxy is True:
            self._proxies = await self.get_proxies()
            if len(self._proxies) == 1:
                proxies = self._proxies[0]
                if not proxies.startswith('http'):
                    proxies = f"http://{proxies}"
            elif len(self._proxies) > 1:
                proxy = random.choice(self._proxies)
                if not proxy.startswith('http'):
                    proxies = f"http://{proxy}"
                else:
                    proxies = proxy
            else:
                self._proxies = None

        # Define custom SSL context
        ssl_context = ssl.create_default_context()
        # Disable older protocols if needed
        ssl_context.options |= ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1
        # Ensure at least TLS 1.2 is used
        ssl_context.minimum_version = ssl.TLSVersion.TLSv1_2
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE

        # Use AsyncHTTPTransport to pass in SSL context explicitly
        transport = httpx.AsyncHTTPTransport(
            retries=num_retries,
            verify=ssl_context
        )
        timeout = httpx.Timeout(
            timeout=timeout,
            connect=connect_timeout,
            read=read_timeout,
            write=write_timeout,
            pool=pool_timeout
        )
        async with httpx.AsyncClient(
            cookies=cookies,
            proxy=proxies or None,
            transport=transport,
            headers=headers,
            timeout=timeout,
            http2=True,
            follow_redirects=True,
            **kwargs
        ) as client:
            try:
                response = await client.get(
                    url,
                    params=params  # Pass query parameters here
                )
                response.raise_for_status()
                return response
            except httpx.TimeoutException:
                print("Request timed out.")
                raise
            except httpx.HTTPError as ex:
                print(f"HTTP error occurred: {ex}")
                raise httpx.HTTPError(ex) from ex
            except Exception as exc:
                print('EXC > ', exc)
                raise ComponentError(
                    f"An error occurred: {exc}"
                ) from exc

    @backoff.on_exception(
        backoff.expo,
        (httpx.HTTPStatusError, httpx.TimeoutException),  # Catch HTTP errors and timeouts
        max_tries=3,
        max_time=120,
        jitter=backoff.full_jitter,
        on_backoff=lambda details: print(
            f"Retrying HTTP Get: attempt {details['tries']} after {details['wait']:0.2f}s"
        ),
    )
    async def _post(
        self,
        url: str,
        cookies: httpx.Cookies,
        params: Dict[str, Any] = None,
        headers: Dict[str, str] = None,
        data: Dict[str, str] = None,
        follow_redirects: bool = True,
        raise_for_status: bool = True,
        use_proxy: bool = True,
        free_proxy: bool = False,
    ) -> Dict[str, Any]:
        """
        Make an asynchronous HTTP POST request, returning the response object.

        Args:
            url (str): The URL to send the POST request to.
            cookies (httpx.Cookies): Cookies to include in the request.
            params (dict): Dictionary of query parameters to include in the URL.

        Returns:
            Response: The response object from the httpx.
        """
        proxies = None
        if use_proxy is True:
            self._proxies = await self.get_proxies()
            if len(self._proxies) == 1:
                proxies = self._proxies[0]
                if not proxies.startswith('http'):
                    proxies = f"http://{proxies}"
            elif len(self._proxies) > 1:
                proxy = random.choice(self._proxies)
                if not proxy.startswith('http'):
                    proxies = f"http://{proxy}"
                else:
                    proxies = proxy
            else:
                self._proxies = None

        # Define custom SSL context
        ssl_context = ssl.create_default_context()
        # Disable older protocols if needed
        ssl_context.options |= ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1
        # Ensure at least TLS 1.2 is used
        ssl_context.minimum_version = ssl.TLSVersion.TLSv1_2
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE

        # Use AsyncHTTPTransport to pass in SSL context explicitly
        transport = httpx.AsyncHTTPTransport(retries=2, verify=ssl_context)
        timeout = httpx.Timeout(connect=5.0, read=20.0, write=5.0, pool=20.0)

        async with httpx.AsyncClient(
            cookies=cookies,
            proxy=proxies or None,
            transport=transport,
            headers=headers,
            timeout=timeout
        ) as client:
            try:
                response = await client.post(
                    url,
                    data=data,
                    params=params,
                    follow_redirects=follow_redirects
                )
                if raise_for_status:
                    response.raise_for_status()
                return response
            except httpx.TimeoutException:
                print("Request timed out.")
                raise
            except httpx.HTTPError as ex:
                print(f"HTTP error occurred: {ex}")
                raise httpx.HTTPError(ex) from ex
            except Exception as exc:
                print('EXC > ', exc)
                raise ComponentError(
                    f"An error occurred: {exc}"
                ) from exc

    @backoff.on_exception(
        backoff.expo,
        (httpx.HTTPStatusError, httpx.TimeoutException),  # Catch HTTP errors and timeouts
        max_tries=3,
        max_time=120,
        jitter=backoff.full_jitter,
        on_backoff=lambda details: print(
            f"Retrying HTTP Get: attempt {details['tries']} after {details['wait']:0.2f}s"
        ),
    )
    async def api_get(
        self,
        url: str,
        cookies: httpx.Cookies = None,
        params: Dict[str, Any] = None,
        headers: Dict[str, str] = None,
        use_proxy: bool = None,
        free_proxy: bool = False,
        use_http2: bool = True
    ) -> Dict[str, Any]:
        """
        Make an asynchronous HTTP GET request.

        Args:
            url (str): The URL to send the GET request to.
            cookies (httpx.Cookies): Cookies to include in the request.
            params (dict): Dictionary of query parameters to include in the URL.

        Returns:
            dict: The JSON response from the API if the request is successful.
            Returns an empty dictionary if the request fails.
        """
        proxies = None
        use_proxy = self.use_proxy if use_proxy is None else use_proxy
        if use_proxy is True:
            self._proxies = await self.get_proxies(free_proxy=free_proxy)
            if len(self._proxies) == 1:
                proxies = self._proxies[0]
                proxies = proxies.strip() if proxies.startswith('http') else f"http://{proxies}"
            elif len(self._proxies) > 1:
                proxy = random.choice(self._proxies)
                if not proxy.startswith('http'):
                    proxies = f"http://{proxy}"
                else:
                    proxies = proxy
            else:
                proxies = None
            self._logger.debug(
                f'SELECTED PROXY: {proxies}'
            )
        # Define custom SSL context
        ssl_context = ssl.create_default_context()
        # Disable older protocols if needed
        ssl_context.options |= ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1
        # Ensure at least TLS 1.2 is used
        ssl_context.minimum_version = ssl.TLSVersion.TLSv1_2
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE

        # Use AsyncHTTPTransport to pass in SSL context explicitly
        transport = httpx.AsyncHTTPTransport(retries=2, verify=ssl_context)
        timeout = httpx.Timeout(connect=5.0, read=20.0, write=5.0, pool=20.0)

        async with httpx.AsyncClient(
            cookies=cookies,
            proxy=proxies or None,
            transport=transport,
            headers=headers,
            timeout=timeout,
            http2=use_http2
        ) as client:
            try:
                response = await client.get(
                    url,
                    params=params
                )
                response.raise_for_status()
                if response.status_code < 400:
                    return response.json()
                else:
                    print(
                        f"API request failed with status code {response.status_code}"
                    )
                return {}
            except httpx.TimeoutException:
                print("Request timed out.")
                raise
            except httpx.HTTPError as ex:
                print(f"HTTP error occurred: {ex}")
                raise httpx.HTTPError(ex) from ex
            except Exception as exc:
                print('EXC > ', exc)
                raise ComponentError(
                    f"An error occurred: {exc}"
                ) from exc
    @backoff.on_exception(
        backoff.expo,
        (httpx.HTTPStatusError, httpx.TimeoutException),  # Catch HTTP errors and timeouts
        max_tries=3,
        max_time=120,
        jitter=backoff.full_jitter,
        on_backoff=lambda details: print(
            f"Retrying HTTP Get: attempt {details['tries']} after {details['wait']:0.2f}s"
        ),
    )
    async def api_post(
        self,
        url: str,
        payload: Dict,
        cookies: httpx.Cookies = None,
        use_proxy: bool = None,
        free_proxy: bool = False,
        full_response: bool = False
    ) -> Dict[str, Any]:
        proxies = None
        use_proxy = self.use_proxy if use_proxy is None else use_proxy
        if use_proxy is True:
            self._proxies = await self.get_proxies(free_proxy=free_proxy)
            if len(self._proxies) == 1:
                proxies = self._proxies[0]
                proxies = proxies.strip() if proxies.startswith('http') else f"http://{proxies}"
            elif len(self._proxies) > 1:
                proxy = random.choice(self._proxies)
                if not proxy.startswith('http'):
                    proxies = f"http://{proxy}"
                else:
                    proxies = proxy
            else:
                proxies = None
        #self._logger.debug(f'SELECTED PROXY: {proxies}')
        # Define custom SSL context
        ssl_context = ssl.create_default_context()
        # Disable older protocols if needed
        ssl_context.options |= ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1
        # Ensure at least TLS 1.2 is used
        ssl_context.minimum_version = ssl.TLSVersion.TLSv1_2
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE

        # Use AsyncHTTPTransport to pass in SSL context explicitly
        transport = httpx.AsyncHTTPTransport(retries=2, verify=ssl_context)
        timeout = httpx.Timeout(connect=5.0, read=20.0, write=5.0, pool=20.0)

        async with httpx.AsyncClient(
            cookies=cookies,
            proxy=proxies or None,
            transport=transport,
            headers=self.headers,
            timeout=timeout
        ) as client:
            try:
                response = await client.post(
                    url,
                    json=payload,
                    headers=self.headers,
                    timeout=timeout
                )
                response.raise_for_status()
                if full_response:
                    return response
                if response.status_code < 400:
                    return response.json()
                else:
                    print(
                        f"API request failed with status code {response.status_code}"
                    )
                return {}
            except httpx.TimeoutException:
                raise
            except (httpx.HTTPError) as ex:
                raise httpx.HTTPError(ex)
            except Exception as exc:
                print('EXC > ', exc)
                raise ComponentError(
                    f"An error occurred: {exc}"
                )

    @backoff.on_exception(
        backoff.expo,
        (RatelimitException, TimeoutException, DuckDuckGoSearchException),
        max_tries=5,
        max_time=120,  # Extended max time to allow sufficient retries
        jitter=backoff.full_jitter,  # Introduces randomization in retry timing
        on_backoff=lambda details: print(
            f"Retrying DuckDuckGo search: attempt {details['tries']} after {details['wait']:0.2f}s"
        ),
    )
    async def _search_duckduckgo(
        self,
        query: str,
        max_results: int = 5,
        use_proxy: bool = True,
        timeout: int = 20,
        headers: dict = None,
        region: str = "wt-wt",
        backend: str = 'lite'
    ):
        """
        Search DuckDuckGo for a given query.

        Args:
            query (str): The search query.
            max_results (int): The maximum number of results to return.
            use_proxy (bool): Whether to use a proxy for the search.

        Returns:
            list: A list of search results.
        """
        proxies = None
        if use_proxy is True:
            self._proxies = await self.get_proxies()
            if len(self._proxies) == 1:
                proxies = self._proxies[0]
                if not proxies.startswith('http'):
                    proxies = f"http://{proxies}"
            elif len(self._proxies) > 1:
                proxy = random.choice(self._proxies)
                if not proxy.startswith('http'):
                    proxies = f"http://{proxy}"
                else:
                    proxies = proxy
            else:
                self._proxies = None
        if headers is None:
            headers = {}
        headers = {**self.headers, **headers}
        headers["User-Agent"] = random.choice(ua)
        try:
            with DDGS(
                headers=headers,
                proxy=proxies,
                timeout=timeout,
                verify=False
            ) as search:
                # 🐵 Monkey Patching Primp Client to avoid Rate-Limits issues:
                search.client = primp.Client(
                    headers=search.headers,
                    proxy=proxies,
                    timeout=timeout,
                    cookie_store=False,  # 🚀 Disable cookie persistence dynamically
                    referer=True,
                    impersonate=random.choice(DDGS._impersonates),
                    impersonate_os=random.choice(DDGS._impersonates_os),
                    follow_redirects=False,
                    verify=False,
                )
                return search.text(
                    keywords=query,
                    timelimit=timeout,
                    max_results=max_results,
                    backend=backend,
                    region=region
                )
        except DuckDuckGoSearchException as e:
            raise RatelimitException(
                f"Error on DuckDuckGo Search: {e}"
            ) from e
        except Exception as e:
            raise RuntimeError(
                f"DuckDuckGo Error: {e}"
            ) from e

    @backoff.on_exception(
        backoff.expo,
        (httpx.HTTPStatusError, httpx.TimeoutException, httpx.RemoteProtocolError),  # Catch HTTP errors and timeouts
        max_tries=5,
        max_time=120,
        jitter=backoff.full_jitter,
        on_backoff=lambda details: print(
            f"Retrying Google Search: attempt {details['tries']} after {details['wait']:0.2f}s"
        ),
    )
    async def _search_google(
        self,
        query: str,
        exact_term: str = None,
        max_results: int = 5,
        use_proxy: bool = True,
        timeout: int = 20,
        headers: dict = None,
        region: str = None,
        country: str = None,
        language: str = None,
        use_primp: bool = False,
        **kwargs
    ):
        if headers:
            headers = {
                **self.headers,
                **headers,
                "Referer": "https://www.google.com/",
            }
        proxies = None
        if use_proxy is True:
            self._proxies = await self.get_proxies()
            if len(self._proxies) == 1:
                proxies = self._proxies[0]
                if not proxies.startswith('http'):
                    proxies = f"http://{proxies}"
            elif len(self._proxies) > 1:
                proxy = random.choice(self._proxies)
                if not proxy.startswith('http'):
                    proxy = f"http://{proxy}"
                proxies = {
                    "http://": httpx.AsyncHTTPTransport(
                        proxy=f"http://{proxy}"
                    ),
                    "https://": httpx.AsyncHTTPTransport(
                        proxy=f"http://{proxy}"
                    ),
                }
            else:
                self._proxies = None
        args = {
            "q": query,
            "cx": str(GOOGLE_SEARCH_ENGINE_ID),
            "num": str(max_results),
            "key": str(self.google_api_key),
            "hl": "en",             # UI language in English
            "sort": "date",         # Prefer newer content
        }
        if region:
            args["gl"] = region  # Geolocation
        if country:
            args["cr"] = country  # Country restriction
        if language:
            args["hl"] = language  # Language preference
        if exact_term:
            args["exactTerms"] = exact_term
        if use_primp:
            # Use Primp Client instead httpx:
            client = primp.Client(
                headers=headers,
                proxy=proxies,  # Use proxy if enabled
                timeout=timeout,
                cookie_store=False,  # 🚀 Disable cookie persistence
                referer=True,
                impersonate=random.choice(impersonates),
                impersonate_os=random.choice(impersonates_os),
                follow_redirects=True,
                verify=False
            )
            try:
                query = quote(query)
                search_url = f"https://cse.google.com/cse?cx={GOOGLE_SEARCH_ENGINE_ID}#gsc.tab=0&gsc.q={query}&gsc.sort="  # noqa
                response = client.get(
                    search_url,
                    **kwargs
                )
                if response.status_code != 200:
                    raise RuntimeError(
                        f"Google Search API failed with status {response.status_code}: {response.text}"
                    )
                return self._parse_google_cse_results(response.text, max_results)
            except Exception as e:
                print(f"Unexpected error: {e}")
                raise RuntimeError(
                    f"Primp Unexpected error: {e}"
                ) from e
        else:
            t = httpx.Timeout(timeout, connect=5.0, read=20.0, write=5.0, pool=20.0)
            async with httpx.AsyncClient(
                proxy=proxies,
                timeout=t,
            ) as client:
                try:
                    response = await client.get(
                        "https://customsearch.googleapis.com/customsearch/v1",
                        headers=headers,
                        params=args,
                        **kwargs
                    )
                    response.raise_for_status()
                    if response.status_code < 400:
                        return response.json()
                    else:
                        raise RuntimeError(
                            f"Google Search API failed: {response.text}, status: {response.status_code}"
                        )
                except httpx.HTTPStatusError as e:
                    print(f"Search Google: HTTP error: {e.response.status_code} - {e.response.text}")
                    raise
                except httpx.TimeoutException:
                    print("Search Google: Request timed out")
                    raise
                except httpx.RemoteProtocolError:  # ✅ Catch server disconnection error
                    print("Search Google: Server disconnected. Retrying with redirect enabled...")
                    raise
                except Exception as e:
                    print(f"Search Google: Unexpected error: {e}")
                    raise

    def get_httpx_cookies(self, domain: str = None, path: str = '/', cookies: dict = None):
        httpx_cookies = httpx.Cookies()
        if cookies is None:
            cookies = {}
        for key, value in cookies.items():
            httpx_cookies.set(
                key, value,
                domain=domain,
                path=path
            )
        return httpx_cookies

    def _parse_google_cse_results(self, html_content: str, max_results: int):
        """
        Extracts search results from the rendered HTML of `cse.google.com/cse`.

        Args:
            html_content (str): The HTML response from the search.
            max_results (int): Max number of results to return.

        Returns:
            list: List of extracted URLs and titles.
        """
        soup = bs(html_content, "html.parser")
        results = []

        print('CONTENT > ', html_content)

        # Extract results from the dynamically loaded content
        for item in soup.select(".gsc-webResult")[:max_results]:  # Adjust this selector if necessary
            title_tag = item.select_one(".gs-title")
            url_tag = item.select_one(".gs-title a")

            if title_tag and url_tag:
                title = title_tag.get_text(strip=True)
                url = url_tag["href"]
                results.append({"title": title, "url": url})

        return results
    @backoff.on_exception(
        backoff.expo,
        (httpx.HTTPStatusError, httpx.TimeoutException),  # Catch HTTP errors and timeouts
        max_tries=3,
        max_time=120,
        jitter=backoff.full_jitter,
        on_backoff=lambda details: logging.warning(
            f"Retrying HTTP Get: attempt {details['tries']} after {details['wait']:0.2f}s"
        ),
        giveup=lambda e: isinstance(e, httpx.HTTPStatusError) and e.response.status_code not in [429, 500, 502, 503, 504]  # pylint: disable=C0301  # noqa
    )
    async def _request(
        self,
        url: str,
        method: str = 'get',
        cookies: Optional[httpx.Cookies] = None,
        params: Optional[Dict[str, Any]] = None,
        data: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: Union[int, float] = 30.0,
        use_proxy: bool = True,
        free_proxy: bool = False,
        use_ssl: bool = True,
        use_json: bool = False,
        follow_redirects: bool = True,
        raise_for_status: bool = True,
        full_response: bool = False,
        connect_timeout: Union[int, float] = 5.0,
        read_timeout: Union[int, float] = 20.0,
        write_timeout: Union[int, float] = 5.0,
        pool_timeout: Union[int, float] = 20.0,
        num_retries: int = 2,
        **kwargs
    ) -> Dict[str, Any]:
        """
        Make an asynchronous HTTPx request, returning the response object.

        Args:
            url (str): The URL to send the request to.
            method (str): The HTTP method to use (default: 'get').
            headers (dict, optional): Dictionary of HTTP headers to include in the request.
            cookies (httpx.Cookies, optional): Cookies to include in the request.
            params (dict, optional): Dictionary of query parameters to include in the URL.
            data (dict, optional): Dictionary of data to send in the request body.
            timeout (float, optional): Total timeout for the request in seconds.
            use_proxy (bool): Whether to use a proxy for the request.
            free_proxy (bool): Whether to use a free proxy.
            use_ssl (bool): Whether to use SSL for the request.
            use_json (bool): Whether to send data as JSON.
            follow_redirects (bool): Whether to follow redirects.
            raise_for_status (bool): Whether to raise an exception for HTTP errors.
            full_response (bool): Whether to return the full response object.
            connect_timeout (float): Timeout for connecting to the server.
            read_timeout (float): Timeout for reading from the server.
            write_timeout (float): Timeout for writing to the server.
            pool_timeout (float): Timeout for connection pool operations.
            num_retries (int): Number of retries to attempt at the transport level.
            **kwargs: Additional arguments to pass to httpx.AsyncClient.

        Returns:
            Tuple[Any, Optional[Dict[str, Any]]]: A tuple containing the result and any error information.

        Raises:
            httpx.TimeoutException: When the request times out.
            httpx.TooManyRedirects: When too many redirects are encountered.
            httpx.HTTPStatusError: When an HTTP error status is encountered.
            httpx.HTTPError: When an HTTP-related error occurs.
            AttributeError: When the HTTP method is invalid.
            RuntimeError: When an unknown error occurs.
        """
        proxies = None
        if use_proxy is True:
            self._proxies = await self.get_proxies()
            if len(self._proxies) == 1:
                proxies = self._proxies[0]
                if not proxies.startswith('http'):
                    proxies = f"http://{proxies}"
            elif len(self._proxies) > 1:
                proxy = random.choice(self._proxies)
                if not proxy.startswith('http'):
                    proxy = f"http://{proxy}"
                proxies = {
                    "http://": httpx.AsyncHTTPTransport(
                        proxy=f"http://{proxy}"
                    ),
                    "https://": httpx.AsyncHTTPTransport(
                        proxy=f"http://{proxy}"
                    ),
                }
            else:
                self._proxies = None

        ssl_context = None
        if use_ssl:
            # Define custom SSL context
            ssl_context = ssl.create_default_context()
            # Disable older protocols if needed
            ssl_context.options |= ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1
            # Ensure at least TLS 1.2 is used
            ssl_context.minimum_version = ssl.TLSVersion.TLSv1_2
            # Make this configurable rather than hardcoded to CERT_NONE
            if kwargs.get('verify_ssl', True):
                ssl_context.check_hostname = True
                ssl_context.verify_mode = ssl.CERT_REQUIRED
            else:
                ssl_context.check_hostname = False
                ssl_context.verify_mode = ssl.CERT_NONE

        # Use AsyncHTTPTransport to pass in SSL context explicitly
        transport_options = {
            'retries': num_retries,
            'verify': ssl_context
        }
        if 'transport_options' in kwargs:
            transport_options.update(kwargs.pop('transport_options'))
        transport = httpx.AsyncHTTPTransport(
            **transport_options
        )
        timeout = httpx.Timeout(
            timeout=timeout,
            connect=connect_timeout,
            read=read_timeout,
            write=write_timeout,
            pool=pool_timeout
        )
        method = method.upper()
        if method not in valid_methods:
            raise ValueError(
                f"Invalid HTTP method: {method}. Must be one of {valid_methods}"
            )
        async with httpx.AsyncClient(
            cookies=cookies,
            proxy=proxies or None,
            transport=transport,
            headers=headers,
            timeout=timeout,
            http2=kwargs.get('use_http2', True),
            follow_redirects=follow_redirects,
            **kwargs
        ) as client:
            try:
                args = {
                    "method": method.upper(),
                    "url": url,
                    "follow_redirects": follow_redirects
                }
                if data:
                    if use_json:
                        args["json"] = data
                    else:
                        args["data"] = data
                if params:
                    args["params"] = params
                response = await client.request(**args)
                if raise_for_status:
                    response.raise_for_status()
                if full_response:
                    if self.logger.isEnabledFor(logging.DEBUG):
                        self.logger.debug(
                            f"Response from {url}: status={response.status_code}, headers={response.headers}"
                        )
                    return response, None
                result, error = await self.process_response(
                    response,
                    url,
                )
                return result, error
            except httpx.TimeoutException:
                self.logger.error("Request timed out.")
                raise
            except httpx.TooManyRedirects:
                self.logger.error("Too many redirects.")
                raise
            except httpx.HTTPStatusError as ex:
                self.logger.error(
                    f"HTTP status error occurred: {ex.response.status_code} - {ex}"
                )
                raise
            except httpx.HTTPError as ex:
                self.logger.error(f"HTTP error occurred: {ex}")
                raise
            except AttributeError as e:
                self.logger.error(f"HTTPx Client doesn't have attribute {method}: {e}")
                raise
            except Exception as exc:
                self.logger.error(f'Unknown Error > {exc}')
                raise RuntimeError(
                    f"An error occurred: {exc}"
                ) from exc



"""External hooks system for AutonomousOrchestrator.

All concrete hook imports are lazy to avoid pulling in heavy
transitive dependencies (asyncpg, watchdog, apscheduler, aioimaplib,
azure-identity, etc.) at package import time.
"""
import importlib
from .base import BaseHook
from .manager import HookManager
from .mixins import HookableAgent
from .models import (
    BrokerHookConfig,
    FileUploadHookConfig,
    FileWatchdogHookConfig,
    HookEvent,
    HookType,
    IMAPHookConfig,
    JiraWebhookConfig,
    MatrixHookConfig,
    MessagingHookConfig,
    PostgresHookConfig,
    SchedulerHookConfig,
    SharePointHookConfig,
    WhatsAppRedisHookConfig,
    create_crew_whatsapp_hook,
    create_multi_agent_whatsapp_hook,
    create_simple_whatsapp_hook,
)


def __getattr__(name: str):
    """Lazy-import concrete hook classes on first access."""
    _lazy_map = {
        # Core hooks
        "SchedulerHook": ".scheduler",
        "FileWatchdogHook": ".file_watchdog",
        "PostgresListenHook": ".postgres",
        "IMAPWatchdogHook": ".imap",
        "JiraWebhookHook": ".jira_webhook",
        "FileUploadHook": ".file_upload",
        "SharePointHook": ".sharepoint",
        # Messaging hooks
        "TelegramHook": ".messaging",
        "WhatsAppHook": ".messaging",
        "MSTeamsHook": ".messaging",
        "WhatsAppRedisHook": ".whatsapp_redis",
        "MatrixHook": ".matrix",
        # Broker hooks
        "BaseBrokerHook": ".brokers.base",
        "RedisBrokerHook": ".brokers.redis",
        "RabbitMQBrokerHook": ".brokers.rabbitmq",
        "MQTTBrokerHook": ".brokers.mqtt",
        "SQSBrokerHook": ".brokers.sqs",
    }
    if name in _lazy_map:
        module = importlib.import_module(_lazy_map[name], package=__name__)
        return getattr(module, name)
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__ = [
    # Core
    "BaseHook",
    "HookManager",
    "HookableAgent",
    "HookEvent",
    "HookType",
    # Hooks (lazy)
    "SchedulerHook",
    "FileWatchdogHook",
    "PostgresListenHook",
    "IMAPWatchdogHook",
    "JiraWebhookHook",
    "FileUploadHook",
    "SharePointHook",
    # Messaging hooks (lazy)
    "TelegramHook",
    "WhatsAppHook",
    "MSTeamsHook",
    "WhatsAppRedisHook",
    "MatrixHook",
    # Brokers (lazy)
    "BaseBrokerHook",
    "RedisBrokerHook",
    "RabbitMQBrokerHook",
    "MQTTBrokerHook",
    "SQSBrokerHook",
    # Configs (eagerly imported — lightweight Pydantic models)
    "SchedulerHookConfig",
    "FileWatchdogHookConfig",
    "PostgresHookConfig",
    "IMAPHookConfig",
    "JiraWebhookConfig",
    "FileUploadHookConfig",
    "BrokerHookConfig",
    "SharePointHookConfig",
    "MessagingHookConfig",
    "WhatsAppRedisHookConfig",
    "MatrixHookConfig",
    # Factory helpers
    "create_simple_whatsapp_hook",
    "create_multi_agent_whatsapp_hook",
    "create_crew_whatsapp_hook",
]



"""
Abstract Transcriber Backend — backward compatibility re-export.

The implementation has moved to `parrot.voice.transcriber.backend`.
This module re-exports for backward compatibility.
"""
from parrot.voice.transcriber.backend import AbstractTranscriberBackend

__all__ = ["AbstractTranscriberBackend"]



"""YAML ontology file loading and validation."""
from __future__ import annotations

import logging
from pathlib import Path
from typing import Any

import yaml
from pydantic import ValidationError

from .schema import OntologyDefinition

logger = logging.getLogger("Parrot.Ontology.Parser")

# Path to package-bundled default ontology files
_DEFAULTS_DIR = Path(__file__).parent / "defaults"


class OntologyParser:
    """Load and validate ontology YAML files against Pydantic schema models.

    Usage::

        parser = OntologyParser()
        definition = parser.load(Path("ontologies/base.ontology.yaml"))
    """

    @staticmethod
    def load(path: Path) -> OntologyDefinition:
        """Load a YAML file and parse it into an OntologyDefinition.

        Args:
            path: Path to the YAML ontology file.

        Returns:
            Validated OntologyDefinition instance.

        Raises:
            FileNotFoundError: If the YAML file does not exist.
            yaml.YAMLError: If the file contains invalid YAML syntax.
            ValidationError: If the YAML content does not match the schema.
        """
        path = Path(path)
        if not path.exists():
            raise FileNotFoundError(f"Ontology file not found: {path}")

        try:
            with open(path, encoding="utf-8") as f:
                raw: dict[str, Any] = yaml.safe_load(f) or {}
        except yaml.YAMLError as e:
            raise yaml.YAMLError(
                f"Invalid YAML in {path}: {e}"
            ) from e

        try:
            definition = OntologyDefinition.model_validate(raw)
        except ValidationError as e:
            logger.error("Ontology validation failed for %s: %s", path, e)
            raise

        logger.debug("Loaded ontology '%s' from %s", definition.name, path)
        return definition

    @staticmethod
    def load_from_dict(data: dict[str, Any]) -> OntologyDefinition:
        """Parse an OntologyDefinition from an already-loaded dict.

        Args:
            data: Dictionary representation of the ontology YAML.

        Returns:
            Validated OntologyDefinition instance.

        Raises:
            ValidationError: If the dict does not match the schema.
        """
        return OntologyDefinition.model_validate(data)

    @staticmethod
    def load_default_base() -> OntologyDefinition:
        """Load the base ontology from package-bundled defaults.

        Returns:
            The base OntologyDefinition shipped with AI-Parrot.

        Raises:
            FileNotFoundError: If the default base file is missing.
        """
        base_path = _DEFAULTS_DIR / "base.ontology.yaml"
        return OntologyParser.load(base_path)

    @staticmethod
    def get_defaults_dir() -> Path:
        """Return the path to the package-bundled defaults directory.

        Returns:
            Path to the defaults/ directory within the ontology package.
        """
        return _DEFAULTS_DIR



from typing import Type, List
import ast
from pydantic import BaseModel, Field
from .base import WebAppGenerator


class StreamlitApp(BaseModel):
    """Pydantic schema for Streamlit application output."""
    code: str = Field(description="Complete Streamlit Python code")
    title: str = Field(description="App title")
    description: str = Field(description="Brief description of the app")
    requirements: List[str] = Field(
        description="Python package requirements",
        default_factory=lambda: ["streamlit"]
    )
    features: List[str] = Field(description="List of main features")


class StreamlitGenerator(WebAppGenerator):
    """Generator for Streamlit applications."""

    def get_system_prompt(self) -> str:
        return """You are an expert Streamlit developer creating production-ready applications.

**User Requirements:**
{user_description}

{additional_requirements}

**Technical Specifications:**
- Streamlit version: 1.28+
- Python 3.10+
- Use st.cache_data for expensive computations
- Implement proper error handling with try-except blocks
- Add loading states with st.spinner()
- Use st.session_state for state management
- Add input validation and clear error messages

**Design Guidelines:**
- Clean, modern UI with st.columns for layout
- Consistent spacing with st.divider() and padding
- Helpful tooltips using help parameter
- Clear section headers with st.header() and st.subheader()
- Data export functionality where appropriate
- Mobile-responsive considerations

**Code Structure:**
1. Required imports at the top
2. Page configuration using st.set_page_config() with appropriate title and layout
3. Helper functions decorated with @st.cache_data where appropriate
4. Main application logic
5. if __name__ == "__main__": block at the end

**Quality Standards:**
- All functions must have docstrings
- Use type hints for function parameters
- Include comments for complex logic
- Handle edge cases and errors gracefully
- Use st.empty() for dynamic content updates
- Add st.success(), st.warning(), st.error() for user feedback

**Example of Expected Quality:**
{high_quality_example}

Generate a complete, production-ready Streamlit application now."""

    def get_output_schema(self) -> Type[BaseModel]:
        return StreamlitApp

    def get_examples(self) -> str:
        return '''
import streamlit as st
import pandas as pd
import plotly.express as px
from datetime import datetime

# Page configuration
st.set_page_config(
    page_title="Data Analysis Dashboard",
    page_icon="📊",
    layout="wide",
    initial_sidebar_state="expanded"
)

# Cache expensive operations
@st.cache_data
def load_data() -> pd.DataFrame:
    """Load and prepare sample data."""
    return pd.DataFrame({
        'date': pd.date_range('2024-01-01', periods=100),
        'category': ['A', 'B', 'C', 'D'] * 25,
        'value': np.random.randint(10, 100, 100)
    })

@st.cache_data
def calculate_metrics(df: pd.DataFrame) -> dict:
    """Calculate summary metrics."""
    return {
        'total': df['value'].sum(),
        'average': df['value'].mean(),
        'max': df['value'].max(),
        'min': df['value'].min()
    }

def main():
    """Main application function."""
    st.title("📊 Data Analysis Dashboard")
    st.markdown("Analyze your data with interactive visualizations")

    # Sidebar controls
    with st.sidebar:
        st.header("Filters")
        categories = st.multiselect(
            "Select Categories",
            options=['A', 'B', 'C', 'D'],
            default=['A', 'B'],
            help="Choose which categories to display"
        )

    # Load data
    try:
        with st.spinner("Loading data..."):
            df = load_data()

        # Filter data
        filtered_df = df[df['category'].isin(categories)]

        if filtered_df.empty:
            st.warning("No data available for selected filters")
            return

        # Metrics row
        metrics = calculate_metrics(filtered_df)
        col1, col2, col3, col4 = st.columns(4)

        with col1:
            st.metric("Total", f"{metrics['total']:,.0f}")
        with col2:
            st.metric("Average", f"{metrics['average']:.1f}")
        with col3:
            st.metric("Maximum", metrics['max'])
        with col4:
            st.metric("Minimum", metrics['min'])

        st.divider()

        # Visualization
        st.subheader("Trend Analysis")
        fig = px.line(
            filtered_df,
            x='date',
            y='value',
            color='category',
            title='Values Over Time'
        )
        st.plotly_chart(fig, use_container_width=True)

        # Data table
        with st.expander("View Raw Data"):
            st.dataframe(filtered_df, use_container_width=True)

            # Export functionality
            csv = filtered_df.to_csv(index=False)
            st.download_button(
                label="Download CSV",
                data=csv,
                file_name=f"data_export_{datetime.now():%Y%m%d}.csv",
                mime="text/csv"
            )

        st.success("✓ Analysis complete!")

    except Exception as e:
        st.error(f"An error occurred: {str(e)}")
        st.exception(e)

if __name__ == "__main__":
    main()
'''

    def validate_output(self, code: str) -> dict:
        """Validate Python syntax of generated Streamlit code."""
        errors = []
        warnings = []

        try:
            ast.parse(code)
        except SyntaxError as e:
            errors.append(f"Syntax error: {e}")

        # Check for required Streamlit imports
        if 'import streamlit' not in code:
            warnings.append("Missing streamlit import")

        # Check for page config
        if 'st.set_page_config' not in code:
            warnings.append("Missing st.set_page_config() - recommended for better UX")

        return {
            "valid": len(errors) == 0,
            "errors": errors,
            "warnings": warnings
        }



"""
Integration Bot Manager.

Manages lifecycle of bots (Telegram, MS Teams, WhatsApp) exposing AI-Parrot agents.
Loads configuration from {ENV_DIR}/integrations_bots.yaml (or telegram_bots.yaml fallback).
"""
import asyncio
from pathlib import Path
from typing import Dict, List, Optional, Tuple, TYPE_CHECKING
import yaml

from aiogram import Bot, Dispatcher
from aiogram.client.default import DefaultBotProperties
from aiogram.enums import ParseMode
from navconfig import BASE_DIR
from navconfig.logging import logging
from ..conf import AGENTS_DIR
from .models import (
    IntegrationBotConfig,
    TelegramAgentConfig,
    MSTeamsAgentConfig,
    WhatsAppAgentConfig,
    SlackAgentConfig,
)
if TYPE_CHECKING:
    from .telegram.wrapper import TelegramAgentWrapper
    from .msteams.wrapper import MSTeamsAgentWrapper
    from .whatsapp.wrapper import WhatsAppAgentWrapper
    from .slack.wrapper import SlackAgentWrapper
    from ..manager import BotManager
    from ..bots.abstract import AbstractBot


ENV_DIR = BASE_DIR.joinpath('env')


class IntegrationBotManager:
    """
    Manages bot integrations for exposed agents.
    
    Supports:
    - Telegram
    - MS Teams
    - WhatsApp
    """

    def __init__(self, bot_manager: 'BotManager'):
        self.bot_manager = bot_manager
        self.logger = logging.getLogger("IntegrationBotManager")

        # Active bots
        self.telegram_bots: Dict[str, Tuple[Bot, Dispatcher, 'TelegramAgentWrapper']] = {}
        self.msteams_bots: Dict[str, 'MSTeamsAgentWrapper'] = {}
        self.whatsapp_bots: Dict[str, 'WhatsAppAgentWrapper'] = {}
        self.slack_bots: Dict[str, 'SlackAgentWrapper'] = {}

        # Matrix crew transport (FEAT-044)
        self.matrix_crew: Optional[object] = None  # MatrixCrewTransport

        self._polling_tasks: List[asyncio.Task] = []
        self._config: Optional[IntegrationBotConfig] = None

    def _get_config_path(self) -> Path:
        """Get path to integrations_bots.yaml (preferred) or telegram_bots.yaml."""
        p = ENV_DIR / "integrations_bots.yaml"
        if p.exists():
            return p
        return ENV_DIR / "telegram_bots.yaml"

    async def load_config(self) -> Optional[IntegrationBotConfig]:
        """Load configuration."""
        config_path = self._get_config_path()
        
        if not config_path.exists():
            self.logger.debug("No integration config found.")
            return None
            
        try:
            with open(config_path, 'r', encoding='utf-8') as f:
                data = yaml.safe_load(f)
                
            if not data:
                return None
                
            # Use the unified config parser
            config = IntegrationBotConfig.from_dict(data)
            
            errors = config.validate()
            if errors:
                for error in errors:
                    self.logger.error(f"Config Error: {error}")
                return None
                
            self._config = config
            return config
            
        except Exception as e:
            self.logger.error(f"Error loading integration config: {e}", exc_info=True)
            return None

    async def _get_agent(self, chatbot_id: str, system_prompt_override: Optional[str] = None) -> Optional['AbstractBot']:
        """Get agent instance from BotManager."""
        agent = await self.bot_manager.get_bot(chatbot_id)
        if not agent:
            self.logger.error(f"Agent '{chatbot_id}' not found.")
            return None
            
        if system_prompt_override and hasattr(agent, 'system_prompt'):
            agent.system_prompt = system_prompt_override
            
        return agent

    async def startup(self, extra_config: Optional[dict] = None) -> None:
        """Start all configured bots.

        Args:
            extra_config: Optional dict with additional integration keys.
                Supports ``"matrix_crew"`` → path to YAML config file for
                ``MatrixCrewTransport`` (FEAT-044).
        """
        self.logger.info("Starting Integration Manager...")

        # Matrix crew transport (FEAT-044)
        if extra_config and "matrix_crew" in extra_config:
            await self._start_matrix_crew(extra_config["matrix_crew"])

        config = await self.load_config()
        if not config:
            return

        for name, agent_config in config.agents.items():
            try:
                if isinstance(agent_config, TelegramAgentConfig):
                    await self._start_telegram_bot(name, agent_config)
                elif isinstance(agent_config, MSTeamsAgentConfig):
                    await self._start_msteams_bot(name, agent_config)
                elif isinstance(agent_config, WhatsAppAgentConfig):
                    await self._start_whatsapp_bot(name, agent_config)
                elif isinstance(agent_config, SlackAgentConfig):
                    await self._start_slack_bot(name, agent_config)
            except Exception as e:
                self.logger.error(f"Failed to start bot {name}: {e}", exc_info=True)

    async def _start_telegram_bot(self, name: str, config: TelegramAgentConfig):
        agent = await self._get_agent(config.chatbot_id, config.system_prompt_override)
        if not agent:
            return

        bot = Bot(token=config.bot_token, default=DefaultBotProperties(parse_mode=ParseMode.MARKDOWN))
        dp = Dispatcher()
        from .telegram.wrapper import TelegramAgentWrapper
        wrapper = TelegramAgentWrapper(agent, bot, config)
        dp.include_router(wrapper.router)
        
        self.telegram_bots[name] = (bot, dp, wrapper)
        
        task = asyncio.create_task(
            self._run_polling(name, dp, bot),
            name=f"telegram_polling_{name}"
        )
        self._polling_tasks.append(task)
        self.logger.info(f"✅ Started Telegram bot '{name}'")

    async def _run_polling(self, name: str, dp: Dispatcher, bot: Bot):
        try:
            await dp.start_polling(
                bot,
                allowed_updates=["message", "callback_query"],
                handle_signals=False,
                close_bot_session=True
            )
        except asyncio.CancelledError:
            pass
        except Exception as e:
            self.logger.error(f"Polling error for {name}: {e}")

    async def _start_msteams_bot(self, name: str, config: MSTeamsAgentConfig):
        agent = await self._get_agent(config.chatbot_id)
        if not agent:
            return
            
        # Initialize Wrapper (which registers the route)
        from .msteams.wrapper import MSTeamsAgentWrapper
        wrapper = MSTeamsAgentWrapper(
            agent=agent,
            config=config,
            app=self.bot_manager.get_app(),
            forms_directory=config.forms_directory or AGENTS_DIR / "forms",
        )
        self.msteams_bots[name] = wrapper
        self.logger.info(f"✅ Started MS Teams bot '{name}'")

    async def _start_whatsapp_bot(self, name: str, config: WhatsAppAgentConfig):
        agent = await self._get_agent(config.chatbot_id, config.system_prompt_override)
        if not agent:
            return

        # Initialize Wrapper (which registers the webhook routes)
        from .whatsapp.wrapper import WhatsAppAgentWrapper
        wrapper = WhatsAppAgentWrapper(
            agent=agent,
            config=config,
            app=self.bot_manager.get_app(),
        )
        self.whatsapp_bots[name] = wrapper
        self.logger.info(f"✅ Started WhatsApp bot '{name}'")


    async def _start_slack_bot(self, name: str, config: SlackAgentConfig):
        # Suppress the verbose DEBUG/INFO output from the slack_sdk internals
        # (heartbeats, raw WebSocket frames, HTTP request dumps, etc.)
        logging.getLogger("slack_sdk").setLevel(logging.WARNING)

        agent = await self._get_agent(config.chatbot_id, config.system_prompt_override if hasattr(config, "system_prompt_override") else None)
        if not agent:
            return

        from .slack.wrapper import SlackAgentWrapper
        wrapper = SlackAgentWrapper(
            agent=agent,
            config=config,
            app=self.bot_manager.get_app(),
        )
        self.slack_bots[name] = wrapper

        # Start the wrapper's background cleanup
        await wrapper.start()

        # Check connection mode
        if config.connection_mode == "socket":
            from .slack.socket_handler import SlackSocketHandler

            handler = SlackSocketHandler(wrapper)
            wrapper._socket_handler = handler
            task = asyncio.create_task(
                handler.start(),
                name=f"slack_socket_{name}",
            )
            self._polling_tasks.append(task)
            self.logger.info(f"✅ Started Slack bot '{name}' (Socket Mode)")
        else:
            self.logger.info(f"✅ Started Slack bot '{name}' (Webhook Mode)")
    async def _start_matrix_crew(self, config_path: str) -> None:
        """Start a Matrix multi-agent crew from a YAML config file.

        Args:
            config_path: Path to the YAML crew configuration file.
        """
        try:
            from .matrix.crew import MatrixCrewTransport

            transport = MatrixCrewTransport.from_yaml(config_path)
            await transport.start()
            self.matrix_crew = transport
            self.logger.info("✅ Started Matrix crew transport from %s", config_path)
        except Exception as exc:
            self.logger.error(
                "Failed to start Matrix crew transport: %s", exc, exc_info=True
            )

    async def shutdown(self) -> None:
        """Shutdown bots."""
        self.logger.info("Shutting down Integration Manager...")
        
        # First, cancel all polling tasks to stop the polling loops
        for task in self._polling_tasks:
            if not task.done():
                task.cancel()
        
        # Wait for all polling tasks to complete (with timeout)
        if self._polling_tasks:
            try:
                await asyncio.wait_for(
                    asyncio.gather(*self._polling_tasks, return_exceptions=True),
                    timeout=5.0
                )
                self.logger.info("All polling tasks cancelled successfully")
            except asyncio.TimeoutError:
                self.logger.warning("Timeout waiting for polling tasks to cancel")
            except Exception as e:
                self.logger.error(f"Error while cancelling polling tasks: {e}")
        
        # Now close bot sessions
        for name, (bot, dp, _) in self.telegram_bots.items():
            try:
                self.logger.debug(f"Closing session for bot '{name}'")
                await bot.session.close()
            except Exception as e:
                self.logger.error(f"Error closing bot session for '{name}': {e}")

        # Stop Slack bots (including Socket Mode handlers)
        for name, wrapper in self.slack_bots.items():
            try:
                self.logger.debug(f"Stopping Slack bot '{name}'")
                # Stop Socket Mode handler if present
                if hasattr(wrapper, "_socket_handler") and wrapper._socket_handler:
                    await wrapper._socket_handler.stop()
                # Stop the wrapper
                await wrapper.stop()
            except Exception as e:
                self.logger.error(f"Error stopping Slack bot '{name}': {e}")
        
        # Stop Matrix crew transport (FEAT-044)
        if self.matrix_crew is not None:
            try:
                await self.matrix_crew.stop()
                self.logger.info("Matrix crew transport stopped")
            except Exception as exc:
                self.logger.error(
                    "Error stopping Matrix crew transport: %s", exc
                )
            self.matrix_crew = None

        # Clear data structures
        self.telegram_bots.clear()
        self.msteams_bots.clear()
        self.whatsapp_bots.clear()
        self.slack_bots.clear()
        self._polling_tasks.clear()

        self.logger.info("Integration Manager shutdown complete")



"""WorkingMemoryToolkit — intermediate result store for analytical operations."""
from .tool import WorkingMemoryToolkit
from .models import (
    OperationType,
    JoinHow,
    AggFunc,
    FilterSpec,
    OperationSpecInput,
    ComputeAndStoreInput,
)

__all__ = [
    "WorkingMemoryToolkit",
    "OperationType",
    "JoinHow",
    "AggFunc",
    "FilterSpec",
    "OperationSpecInput",
    "ComputeAndStoreInput",
]



"""
BigQuery Agent Implementation for AI-Parrot.

Concrete implementation of AbstractDBAgent for Google BigQuery
with support for SQL query language and BigQuery-specific features.
"""

from typing import Dict, Any, List, Optional, Union
import asyncio
from datetime import datetime
from pydantic import Field
from google.cloud import bigquery as bq
from google.oauth2 import service_account
from google.cloud.exceptions import NotFound
from navconfig import config

from .abstract import (
    AbstractDBAgent,
    DatabaseSchema,
    TableMetadata,
)
from ...tools.abstract import AbstractTool, ToolResult, AbstractToolArgsSchema


class BigQueryQueryExecutionArgs(AbstractToolArgsSchema):
    """Arguments for BigQuery query execution."""
    query: str = Field(description="SQL query to execute")
    limit: Optional[int] = Field(
        default=1000, description="Maximum number of rows to return"
    )
    use_legacy_sql: bool = Field(
        default=False, description="Use legacy SQL instead of standard SQL"
    )


class DatasetMetadata:
    """Metadata for BigQuery datasets."""
    def __init__(
        self,
        dataset_id: str,
        project_id: str,
        location: str,
        tables: List[str],
        description: str = None
    ):
        self.dataset_id = dataset_id
        self.project_id = project_id
        self.location = location
        self.tables = tables
        self.description = description


class BigQueryAgent(AbstractDBAgent):
    """
    BigQuery Agent for data warehouse introspection and SQL query generation.

    Supports Google BigQuery with standard SQL and BigQuery-specific features.
    """

    def __init__(
        self,
        name: str = "BigQueryAgent",
        project_id: str = None,
        credentials_file: str = None,
        credentials: Union[str, Dict[str, Any]] = None,
        dataset: str = None,
        location: str = "US",
        max_sample_rows: int = 10,
        **kwargs
    ):
        """
        Initialize BigQuery Agent.

        Args:
            name: Agent name
            project_id: Google Cloud project ID
            credentials_file: Path to service account credentials JSON file
            credentials: Credentials dict or connection string (overrides individual params)
            dataset: Default dataset name
            location: BigQuery location/region (default: US)
            max_sample_rows: Maximum sample rows per table
        """
        self.project_id = project_id
        self.credentials_file = credentials_file
        self.dataset = dataset
        self.location = location
        self.max_sample_rows = max_sample_rows
        self.client: Optional[bq.Client] = None
        self.credentials_obj = None
        self.datasets_cache: Dict[str, DatasetMetadata] = {}

        # Get default credentials if not provided
        if not credentials and not all([project_id, credentials_file]):
            credentials = self._get_default_credentials()

        super().__init__(
            name=name,
            credentials=credentials,
            schema_name=dataset,
            **kwargs
        )

        # Add BigQuery-specific tools
        self._setup_bigquery_tools()

    def _get_default_credentials(self) -> Dict[str, Any]:
        """Get default credentials from config (similar to DatabaseQueryTool)."""
        return {
            'credentials_file': config.get('GOOGLE_APPLICATION_CREDENTIALS'),
            'project_id': config.get('GOOGLE_CLOUD_PROJECT'),
        }

    def _setup_bigquery_tools(self):
        """Setup BigQuery-specific tools."""
        # Add query execution tool
        query_execution_tool = BigQueryQueryExecutionTool(agent=self)
        self.tool_manager.register_tool(query_execution_tool)

        # Add dataset exploration tool
        dataset_exploration_tool = DatasetExplorationTool(agent=self)
        self.tool_manager.register_tool(dataset_exploration_tool)

    async def connect_database(self) -> None:
        """Connect to BigQuery using service account credentials."""
        # Parse credentials
        if isinstance(self.credentials, dict):
            creds = self.credentials.copy()
            self.project_id = creds.get('project_id', self.project_id)
            self.credentials_file = creds.get('credentials_file', self.credentials_file)
            self.dataset = creds.get('dataset', self.dataset)

        try:
            # Initialize BigQuery client
            if self.credentials_file:
                self.credentials_obj = service_account.Credentials.from_service_account_file(
                    self.credentials_file
                )
                if not self.project_id:
                    self.project_id = self.credentials_obj.project_id
                
                self.client = bq.Client(
                    credentials=self.credentials_obj,
                    project=self.project_id
                )
            else:
                # Use application default credentials
                if not self.project_id:
                    raise ValueError("BigQuery project_id is required")
                self.client = bq.Client(project=self.project_id)

            # Test connection by listing datasets (limit to 1)
            datasets = list(self.client.list_datasets(max_results=1))
            
            self.logger.info(
                f"Successfully connected to BigQuery project: {self.project_id}"
            )

        except Exception as e:
            self.logger.error(f"Failed to connect to BigQuery: {e}")
            raise

    async def extract_schema_metadata(self) -> DatabaseSchema:
        """Extract schema metadata from BigQuery (datasets, tables, columns)."""
        if not self.client:
            await self.connect_database()

        try:
            # Get datasets to analyze
            datasets_to_analyze = []
            if self.dataset:
                datasets_to_analyze = [self.dataset]
            else:
                # List all datasets in the project
                datasets = self.client.list_datasets()
                datasets_to_analyze = [dataset.dataset_id for dataset in datasets]

            # Extract tables from each dataset
            all_tables = []
            for dataset_id in datasets_to_analyze:
                tables = await self._extract_tables_from_dataset(dataset_id)
                all_tables.extend(tables)

            schema_metadata = DatabaseSchema(
                database_name=self.project_id,
                database_type="bigquery",
                tables=all_tables,
                views=[],  # Views are included in tables
                functions=[],
                procedures=[],
                metadata={
                    "datasets_analyzed": datasets_to_analyze,
                    "total_tables": len(all_tables),
                    "extraction_timestamp": datetime.now().isoformat(),
                    "location": self.location
                }
            )

            self.logger.info(
                f"Extracted metadata for {len(all_tables)} tables from {len(datasets_to_analyze)} datasets"
            )

            return schema_metadata

        except Exception as e:
            self.logger.error(f"Failed to extract BigQuery schema metadata: {e}")
            raise

    async def _extract_tables_from_dataset(
        self,
        dataset_id: str
    ) -> List[TableMetadata]:
        """Extract all tables from a specific dataset."""
        try:
            dataset_ref = f"{self.project_id}.{dataset_id}"
            tables = self.client.list_tables(dataset_ref)

            table_metadata_list = []
            for table in tables:
                # Get full table details
                table_ref = self.client.get_table(table.reference)
                
                # Extract table metadata
                table_metadata = await self._extract_table_metadata(
                    dataset_id, table_ref
                )
                table_metadata_list.append(table_metadata)

            return table_metadata_list

        except NotFound:
            self.logger.warning(f"Dataset {dataset_id} not found")
            return []
        except Exception as e:
            self.logger.warning(f"Could not extract tables from dataset {dataset_id}: {e}")
            return []

    async def _extract_table_metadata(
        self,
        dataset_id: str,
        table: bq.Table
    ) -> TableMetadata:
        """Extract detailed metadata for a specific table."""
        try:
            # Extract columns
            columns = []
            for field in table.schema:
                columns.append({
                    "name": field.name,
                    "type": field.field_type,
                    "nullable": field.mode != "REQUIRED",
                    "description": field.description or ""
                })

            # Get sample data
            sample_data = []
            try:
                query = f"""
                    SELECT * 
                    FROM `{self.project_id}.{dataset_id}.{table.table_id}` 
                    LIMIT {self.max_sample_rows}
                """
                query_job = self.client.query(query)
                results = query_job.result()
                
                for row in results:
                    sample_data.append(dict(row))
                    
            except Exception as e:
                self.logger.debug(f"Could not get sample data for {table.table_id}: {e}")

            # Determine if it's a view
            is_view = table.table_type == "VIEW"

            return TableMetadata(
                name=table.table_id,
                schema=dataset_id,
                columns=columns,
                primary_keys=[],  # BigQuery doesn't enforce primary keys
                foreign_keys=[],  # BigQuery doesn't enforce foreign keys
                indexes=[],  # BigQuery handles indexing automatically
                description=table.description or f"BigQuery {'view' if is_view else 'table'} in dataset {dataset_id}",
                sample_data=sample_data
            )

        except Exception as e:
            self.logger.warning(
                f"Could not extract metadata for table {table.table_id}: {e}"
            )
            return TableMetadata(
                name=table.table_id,
                schema=dataset_id,
                columns=[],
                primary_keys=[],
                foreign_keys=[],
                indexes=[],
                description="",
                sample_data=[]
            )

    async def generate_query(
        self,
        natural_language_query: str,
        target_tables: Optional[List[str]] = None,
        query_type: str = "SELECT"
    ) -> Dict[str, Any]:
        """Generate BigQuery SQL from natural language."""
        try:
            # Get schema context for the query
            schema_context = []
            if self.schema_metadata:
                # Filter tables based on target_tables
                tables_to_use = self.schema_metadata.tables
                if target_tables:
                    tables_to_use = [
                        t for t in tables_to_use if t.name in target_tables
                    ]
                
                for table in tables_to_use[:5]:  # Limit to top 5
                    schema_context.append({
                        'table': f"{table.schema}.{table.name}",
                        'columns': [
                            {'name': col['name'], 'type': col['type']}
                            for col in table.columns
                        ],
                        'sample_count': len(table.sample_data)
                    })

            # Build SQL query generation prompt
            prompt = self._build_sql_query_prompt(
                natural_language_query=natural_language_query,
                schema_context=schema_context
            )

            # Generate query using LLM
            async with self._llm as client:
                response = await client.ask(
                    prompt=prompt,
                    model=self._llm_model,
                    temperature=0.1
                )

            # Extract SQL query from response
            sql_query = self._extract_sql_from_response(response.output or response.response)

            result = {
                "query": sql_query,
                "query_type": "sql",
                "tables_used": target_tables or [],
                "schema_context_used": len(schema_context),
                "natural_language_input": natural_language_query
            }

            return result

        except Exception as e:
            self.logger.error(f"Failed to generate SQL query: {e}")
            raise

    def _build_sql_query_prompt(
        self,
        natural_language_query: str,
        schema_context: List[Dict[str, Any]]
    ) -> str:
        """Build prompt for SQL query generation."""
        prompt = f"""
You are an expert BigQuery SQL developer.
Generate a BigQuery SQL query based on the natural language request and the provided schema information.

Natural Language Request: {natural_language_query}

Available Tables and Schema:
"""
        for i, context in enumerate(schema_context, 1):
            prompt += f"\n{i}. Table: {context['table']}\n"
            prompt += "   Columns:\n"
            for col in context['columns'][:10]:  # Limit columns shown
                prompt += f"     - {col['name']} ({col['type']})\n"

        prompt += f"""

BigQuery SQL Guidelines:
1. Use standard SQL syntax (not legacy SQL)
2. Fully qualify table names: `project.dataset.table`
3. Use backticks for table/column names with special characters
4. Common BigQuery functions: STRING_AGG, ARRAY_AGG, APPROX_COUNT_DISTINCT
5. Use STRUCT and ARRAY types when appropriate
6. For date/time: CURRENT_TIMESTAMP(), DATE_SUB(), TIMESTAMP_DIFF()
7. Use WITH clauses (CTEs) for complex queries
8. Return only the SQL query without explanations or markdown formatting

Project ID: {self.project_id}

SQL Query:"""

        return prompt

    def _extract_sql_from_response(self, response_text: str) -> str:
        """Extract SQL query from LLM response."""
        # Remove markdown code blocks if present
        if "```" in response_text:
            lines = response_text.split('\n')
            sql_lines = []
            in_code_block = False

            for line in lines:
                if line.strip().startswith("```"):
                    in_code_block = not in_code_block
                    continue
                elif in_code_block:
                    sql_lines.append(line)

            return '\n'.join(sql_lines).strip()
        else:
            return response_text.strip()

    async def execute_query(
        self,
        query: str,
        limit: Optional[int] = 1000
    ) -> Dict[str, Any]:
        """Execute SQL query against BigQuery."""
        try:
            if not self.client:
                await self.connect_database()

            # Add limit if not present
            query_upper = query.upper().strip()
            if limit and 'LIMIT' not in query_upper:
                query = f"{query.rstrip(';')} LIMIT {limit}"

            # Configure query job
            job_config = bq.QueryJobConfig()
            job_config.use_legacy_sql = False

            # Execute query
            query_job = self.client.query(query, job_config=job_config)
            results = query_job.result()

            # Process results
            rows = []
            columns = []
            
            if results.schema:
                columns = [field.name for field in results.schema]
                
            for row in results:
                rows.append(dict(row))

            return {
                "success": True,
                "data": rows,
                "columns": columns,
                "record_count": len(rows),
                "query": query,
                "total_bytes_processed": query_job.total_bytes_processed,
                "total_bytes_billed": query_job.total_bytes_billed
            }

        except Exception as e:
            self.logger.error(f"BigQuery query execution failed: {e}")
            return {
                "success": False,
                "error": str(e),
                "query": query
            }

    async def close(self):
        """Close BigQuery client connection."""
        if self.client:
            self.client.close()


class BigQueryQueryExecutionTool(AbstractTool):
    """Tool for executing SQL queries against BigQuery."""

    name = "execute_bigquery_query"
    description = "Execute SQL queries against Google BigQuery"
    args_schema = BigQueryQueryExecutionArgs

    def __init__(self, agent: BigQueryAgent, **kwargs):
        super().__init__(**kwargs)
        self.agent = agent

    async def _execute(
        self,
        query: str,
        limit: Optional[int] = 1000,
        use_legacy_sql: bool = False
    ) -> ToolResult:
        """Execute BigQuery query."""
        try:
            result = await self.agent.execute_query(query, limit)

            return ToolResult(
                status="success" if result["success"] else "error",
                result=result,
                error=result.get("error"),
                metadata={
                    "query": query,
                    "limit": limit,
                    "bytes_processed": result.get("total_bytes_processed"),
                    "bytes_billed": result.get("total_bytes_billed")
                }
            )

        except Exception as e:
            return ToolResult(
                status="error",
                result=None,
                error=str(e),
                metadata={"query": query}
            )


class DatasetExplorationTool(AbstractTool):
    """Tool for exploring BigQuery datasets and tables."""

    name = "explore_bigquery_datasets"
    description = "Explore available datasets and tables in BigQuery"

    class ExplorationArgs(AbstractToolArgsSchema):
        """Exploration arguments schema."""
        dataset: Optional[str] = Field(
            default=None, description="Specific dataset to explore"
        )
        table: Optional[str] = Field(
            default=None, description="Specific table to explore"
        )
        show_sample_data: bool = Field(
            default=True, description="Include sample data in results"
        )

    args_schema = ExplorationArgs

    def __init__(self, agent: BigQueryAgent, **kwargs):
        super().__init__(**kwargs)
        self.agent = agent

    async def _execute(
        self,
        dataset: Optional[str] = None,
        table: Optional[str] = None,
        show_sample_data: bool = True
    ) -> ToolResult:
        """Explore datasets/tables in BigQuery."""
        try:
            if not self.agent.schema_metadata:
                await self.agent.extract_schema_metadata()

            exploration_result = {
                "datasets": [],
                "tables": [],
                "total_tables": 0
            }

            # Filter tables
            tables_to_explore = self.agent.schema_metadata.tables
            if dataset:
                tables_to_explore = [t for t in tables_to_explore if t.schema == dataset]
            if table:
                tables_to_explore = [t for t in tables_to_explore if t.name == table]

            # Get unique datasets
            datasets = list(set(t.schema for t in tables_to_explore))
            exploration_result["datasets"] = datasets

            # Build table information
            for tbl in tables_to_explore:
                table_info = {
                    "name": tbl.name,
                    "dataset": tbl.schema,
                    "full_name": f"{tbl.schema}.{tbl.name}",
                    "columns": [
                        {'name': col['name'], 'type': col['type']}
                        for col in tbl.columns
                    ],
                    "description": tbl.description
                }

                if show_sample_data and tbl.sample_data:
                    table_info["sample_data"] = tbl.sample_data[:3]

                exploration_result["tables"].append(table_info)

            exploration_result["total_tables"] = len(exploration_result["tables"])

            return ToolResult(
                status="success",
                result=exploration_result,
                metadata={
                    "project": self.agent.project_id,
                    "dataset_filter": dataset,
                    "table_filter": table
                }
            )

        except Exception as e:
            return ToolResult(
                status="error",
                result=None,
                error=str(e)
            )



"""Backward-compatibility shim — hooks have moved to parrot.core.hooks.

This module re-exports everything from the canonical location so that any
existing code importing from ``parrot.autonomous.hooks`` continues to work
without changes.
"""
from parrot.core.hooks import *  # noqa: F401, F403
from parrot.core.hooks import __all__  # noqa: F401



"""Configuration models for TelegramCrewTransport.

Pydantic v2 models for configuring a multi-agent crew
in a Telegram supergroup.
"""
import logging
import os
import re
from typing import Dict, List, Optional

import yaml
from pydantic import BaseModel, Field, field_validator

logger = logging.getLogger(__name__)

_ENV_VAR_PATTERN = re.compile(r"\$\{([^}]+)\}")


def _substitute_env_vars(value: str) -> str:
    """Replace ${ENV_VAR} patterns with environment variable values."""
    def _replace(match: re.Match) -> str:
        var_name = match.group(1)
        env_value = os.environ.get(var_name, "")
        if not env_value:
            logger.warning("Environment variable %s is not set", var_name)
        return env_value
    return _ENV_VAR_PATTERN.sub(_replace, value)


def _walk_and_substitute(obj):
    """Recursively walk a data structure and substitute env vars in strings."""
    if isinstance(obj, str):
        return _substitute_env_vars(obj)
    if isinstance(obj, dict):
        return {k: _walk_and_substitute(v) for k, v in obj.items()}
    if isinstance(obj, list):
        return [_walk_and_substitute(item) for item in obj]
    return obj


class CrewAgentEntry(BaseModel):
    """Configuration for a single agent in the crew."""

    chatbot_id: str = Field(..., description="ID used to retrieve agent from BotManager")
    bot_token: str = Field(..., description="Telegram Bot API token")
    username: str = Field(..., description="Telegram bot username (without @)")
    skills: List[dict] = Field(default_factory=list, description="Agent skill descriptors")
    tags: List[str] = Field(default_factory=list, description="Classification tags")
    accepts_files: List[str] = Field(
        default_factory=list, description="File types the agent can receive"
    )
    emits_files: List[str] = Field(
        default_factory=list, description="File types the agent can produce"
    )
    system_prompt_override: Optional[str] = Field(
        default=None, description="Override the agent's default system prompt"
    )


class TelegramCrewConfig(BaseModel):
    """Root configuration for a multi-agent crew in a Telegram supergroup."""

    group_id: int = Field(..., description="Telegram supergroup chat ID")
    coordinator_token: str = Field(..., description="Bot token for the coordinator bot")
    coordinator_username: str = Field(
        ..., description="Telegram username of the coordinator bot"
    )
    hitl_user_ids: List[int] = Field(
        default_factory=list,
        description="Telegram user IDs of human operators",
    )
    agents: Dict[str, CrewAgentEntry] = Field(
        default_factory=dict,
        description="Map of agent name to agent configuration",
    )
    announce_on_join: bool = Field(
        default=True, description="Announce when an agent joins the crew"
    )
    update_pinned_registry: bool = Field(
        default=True, description="Maintain a pinned registry message"
    )
    reply_to_sender: bool = Field(
        default=True, description="Always @mention the sender in replies"
    )
    silent_tool_calls: bool = Field(
        default=True, description="Suppress tool call messages in the group"
    )
    typing_indicator: bool = Field(
        default=True, description="Show typing indicator while processing"
    )
    max_message_length: int = Field(
        default=4000, description="Max chars per message before chunking"
    )
    temp_dir: str = Field(
        default="/tmp/parrot_crew", description="Temp directory for file exchange"
    )
    max_file_size_mb: int = Field(
        default=50, description="Max file size in MB for document exchange"
    )
    allowed_mime_types: List[str] = Field(
        default=[
            "text/csv",
            "application/json",
            "text/plain",
            "image/png",
            "image/jpeg",
            "application/pdf",
            "application/vnd.apache.parquet",
        ],
        description="Allowed MIME types for document exchange",
    )

    @field_validator("max_message_length")
    @classmethod
    def cap_message_length(cls, v: int) -> int:
        """Telegram messages are capped at 4096 characters."""
        if v > 4096:
            logger.warning(
                "max_message_length %d exceeds Telegram limit, capping to 4096", v
            )
            return 4096
        if v < 1:
            raise ValueError("max_message_length must be positive")
        return v

    @classmethod
    def from_yaml(cls, path: str) -> "TelegramCrewConfig":
        """Load configuration from a YAML file with ${ENV_VAR} substitution.

        Args:
            path: Path to the YAML configuration file.

        Returns:
            A validated TelegramCrewConfig instance.

        Raises:
            FileNotFoundError: If the YAML file does not exist.
            yaml.YAMLError: If the YAML is malformed.
            ValidationError: If the data fails Pydantic validation.
        """
        logger.info("Loading crew config from %s", path)
        with open(path, "r", encoding="utf-8") as f:
            raw = yaml.safe_load(f)
        if raw is None:
            raise ValueError(f"Empty YAML file: {path}")
        data = _walk_and_substitute(raw)
        return cls(**data)



from typing import Any, Optional, Tuple, Dict, List
import contextlib
import io
import base64
import uuid
from .chart import BaseChart
from . import register_renderer
from ...models.outputs import OutputMode


SEABORN_SYSTEM_PROMPT = """SEABORN CHART OUTPUT MODE:
Generate polished statistical visualizations using Seaborn.

REQUIREMENTS:
1. Return Python code in a markdown code block (```python)
2. Import seaborn as sns and set a theme with sns.set_theme()
3. Load or create data directly in the example (use sns.load_dataset or inline data)
4. Store the figure in 'fig' (sns.relplot returns a FacetGrid; use .fig) or fall back to plt.gcf()
5. Add descriptive titles, axis labels, and legend/annotation cues
6. Prefer seaborn high-level functions (relplot, catplot, histplot, heatmap, etc.)
7. Keep charts self-contained—no external files or plt.show()

EXAMPLE:
```python
# Import seaborn
import seaborn as sns

# Apply the default theme
sns.set_theme()

# Load an example dataset
tips = sns.load_dataset("tips")

# Create a visualization
sns.relplot(
    data=tips,
    x="total_bill", y="tip", col="time",
    hue="smoker", style="smoker", size="size",
)
```

Explanation:
- sns.set_theme() ensures a consistent, modern aesthetic.
- Inline dataset loading keeps the code runnable anywhere.
- relplot showcases multi-faceted Seaborn features (faceting, hue, style, size).
- Returning only the code snippet allows the renderer to execute it safely.
"""


@register_renderer(OutputMode.SEABORN, system_prompt=SEABORN_SYSTEM_PROMPT)
class SeabornRenderer(BaseChart):
    """Renderer for Seaborn charts (rendered as static images)."""

    def execute_code(
        self,
        code: str,
        pandas_tool: Any = None,
        execution_state: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Tuple[Any, Optional[str]]:
        """Execute Seaborn code and return all underlying Matplotlib figures."""
        manual_backend = pandas_tool is None
        extra_namespace = None

        if manual_backend:
            import matplotlib
            matplotlib.use('Agg')
            import matplotlib.pyplot as plt
            import seaborn as sns
            extra_namespace = {
                'sns': sns,
                'plt': plt,
                'matplotlib': matplotlib,
            }

        try:
            # Execute using BaseRenderer logic
            context, error = super().execute_code(
                code,
                pandas_tool=pandas_tool,
                execution_state=execution_state,
                extra_namespace=extra_namespace,
                **kwargs,
            )

            if error:
                return None, error

            if not context:
                return None, "Execution context was empty"

            # Find all figure objects
            figures = self._find_chart_objects(context)

            if figures:
                return figures, None

            return None, "Code must define a figure variable (fig, chart, plot) or create matplotlib figures"

        finally:
            if manual_backend:
                with contextlib.suppress(Exception):
                    import matplotlib.pyplot as plt
                    plt.close('all')

    @staticmethod
    def _find_chart_objects(context: Dict[str, Any]) -> List[Any]:
        """Locate all matplotlib figure objects in the local namespace."""
        figures: List[Any] = []
        seen_ids = set()

        def add_fig(obj: Any) -> None:
            if obj is None:
                return

            # Skip renderer / BaseChart instances (like `self`)
            if isinstance(obj, BaseChart):
                return

            # Check if it's a matplotlib Figure
            has_savefig = hasattr(obj, 'savefig')
            has_axes = hasattr(obj, 'axes')

            # Check if it's a Seaborn FacetGrid or similar
            has_fig_attr = hasattr(obj, 'fig') and hasattr(obj.fig, 'savefig')

            # Check if it's a matplotlib Axes
            is_axes = hasattr(obj, 'figure') and hasattr(obj.figure, 'savefig')

            if has_fig_attr:
                # Handle FacetGrid, PairGrid, etc.
                fig = obj.fig
                if id(fig) not in seen_ids:
                    figures.append(fig)
                    seen_ids.add(id(fig))
            elif is_axes:
                # Handle Axes objects
                fig = obj.figure
                if id(fig) not in seen_ids:
                    figures.append(fig)
                    seen_ids.add(id(fig))
            elif has_savefig and has_axes and id(obj) not in seen_ids:
                # Handle Figure objects directly
                figures.append(obj)
                seen_ids.add(id(obj))

        # 1. Priority search for common variable names to preserve order
        priority_vars = ['fig', 'figure', 'chart', 'plot', 'g', 'grid', 'ax', 'axes']
        for var_name in priority_vars:
            if var_name in context:
                add_fig(context[var_name])

        # 2. Scan all locals for other figure objects
        for var_name, obj in context.items():
            if var_name.startswith('_') or var_name in priority_vars:
                continue
            add_fig(obj)

        # 3. Fallback: try to get current figure from plt if available
        if not figures and 'plt' in context:
            try:
                fig = context['plt'].gcf()
                if fig and hasattr(fig, 'savefig') and id(fig) not in seen_ids:
                    figures.append(fig)
            except Exception:
                pass

        return figures

    def _render_chart_content(self, chart_objs: Any, **kwargs) -> str:
        """
        Render Seaborn chart(s) as embedded base64 image(s).
        Handles a single figure or a list of figures.
        """
        # Ensure we have a list
        figures = chart_objs if isinstance(chart_objs, list) else [chart_objs]

        html_parts = []
        img_format = kwargs.get('format', 'png')
        dpi = kwargs.get('dpi', 110)

        for i, chart_obj in enumerate(figures):
            img_id = f"seaborn-chart-{uuid.uuid4().hex[:8]}"

            # Render figure to base64
            buffer = io.BytesIO()
            chart_obj.savefig(buffer, format=img_format, dpi=dpi, bbox_inches='tight')
            buffer.seek(0)
            img_base64 = base64.b64encode(buffer.read()).decode('utf-8')
            buffer.close()

            chart_html = f'''
            <div class="seaborn-chart-wrapper" style="margin-bottom: 20px; width: 100%;">
                <img id="{img_id}"
                     src="data:image/{img_format};base64,{img_base64}"
                     style="width: 100%; height: auto; display: block; margin: 0 auto;"
                     alt="Seaborn Chart {i+1}" />
            </div>
            '''
            html_parts.append(chart_html)

        return "\n".join(html_parts)

    def to_html(
        self,
        chart_obj: Any,
        mode: str = 'partial',
        **kwargs
    ) -> str:
        """Convert Seaborn chart(s) to HTML."""
        kwargs['extra_head'] = kwargs.get('extra_head', '')

        # Call parent to_html (which calls _render_chart_content)
        return super().to_html(chart_obj, mode=mode, **kwargs)

    def to_json(self, chart_obj: Any) -> Optional[Any]:
        """Return metadata noting Seaborn renders as static images."""
        figures = chart_obj if isinstance(chart_obj, list) else [chart_obj]

        results = []
        for fig in figures:
            results.append({
                'type': 'seaborn',
                'note': 'Seaborn visualizations render as Matplotlib figures encoded into base64 images.',
                'figure_size': list(fig.get_size_inches()) if hasattr(fig, 'get_size_inches') else None,
                'dpi': fig.dpi if hasattr(fig, 'dpi') else None,
            })

        return results if len(results) > 1 else results[0] if results else None

    async def render(
        self,
        response: Any,
        theme: str = 'monokai',
        environment: str = 'html',
        include_code: bool = False,
        html_mode: str = 'partial',
        **kwargs
    ) -> Tuple[Any, Optional[Any]]:
        """Render Seaborn chart(s)."""

        # 1. Extract Code
        code = getattr(response, 'code', None)
        output_format = kwargs.get('output_format', environment)

        # Fallback to extracting from text content
        if not code:
            content = self._get_content(response)
            code = self._extract_code(content)

        if not code:
            error_msg = "No chart code found in response"
            if output_format == 'terminal':
                return error_msg, None
            return self._wrap_for_environment(
                f"<div class='error'>{error_msg}</div>",
                output_format
            ), None

        # 2. Execute Code
        chart_objs, error = self.execute_code(
            code,
            pandas_tool=kwargs.pop('pandas_tool', None),
            execution_state=kwargs.pop('execution_state', None),
            **kwargs,
        )

        if error:
            if output_format == 'terminal':
                return f"Error generating chart: {error}", None
            return self._wrap_for_environment(
                self._render_error(error, code, theme),
                output_format
            ), None

        # 3. Handle Jupyter/Notebook Environment
        if output_format in {'jupyter', 'notebook', 'ipython', 'colab'}:
            # For Jupyter, return the figure object(s) directly
            # The frontend will handle rendering them
            if isinstance(chart_objs, list):
                # If multiple figures, return them as a tuple
                return code, chart_objs if len(chart_objs) > 1 else chart_objs[0]
            return code, chart_objs

        # 4. Generate HTML for Web/Terminal
        html_output = self.to_html(
            chart_objs,
            mode=html_mode,
            include_code=include_code,
            code=code,
            theme=theme,
            title=kwargs.get('title', 'Seaborn Chart'),
            icon='🎨',
            dpi=kwargs.get('dpi', 110),
            format=kwargs.get('img_format', 'png'),
            **kwargs
        )

        # 5. Return based on output format
        if output_format == 'html':
            return code, html_output
        elif output_format == 'json':
            return code, self.to_json(chart_objs)
        elif output_format == 'terminal':
            # For terminal, could save to file like Plotly does
            # For now, return the HTML
            return code, html_output

        # Default behavior: Return code as content, HTML as wrapped
        return code, html_output
